/***************************************************************
 * csutils.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <signal.h>
#include <time.h>
#ifndef __SPU__
#include <math.h>
#endif
#include "timebase.h"
#include "cs.h"

void
cs_nsleep(long nsec) {
  int rc;
  struct timespec timespec, rem;

  timespec.tv_sec = (time_t)0;
  timespec.tv_nsec = nsec;
  do {
    rc = nanosleep(&timespec, &rem);
    if(rc != 0) {
      if(errno == EINTR) {
      	timespec = rem;
      	continue;
      }
      perror("cs_nsleep");
      break;
    }
  } while(rc != 0);
  return;
}

void
cs_usleep(int usec) {
  cs_nsleep(usec * 1000);
}

void
cs_exit_perror(const char *msg) {
  perror(msg);
  exit(EXIT_FAILURE);
}

uint32_t
cs_timestamp(void) {
#if defined(__SPU__)
  return (uint32_t)START_CLOCK() / (TIMEBASE / 1000000);
#elif 0 && defined(__PPU__)
  return (uint32_t)get_timebase_lower32() / (TIMEBASE / 1000000);
#else
  struct timeval t;
  while(gettimeofday(&t, NULL) < 0) {
if(2) perror("timestamp: gettimeofday");
    cs_usleep(1);
  }
  return (uint32_t)t.tv_usec;
#endif
}

int32_t
cs_timesub(uint32_t x, uint32_t y) {
  int32_t i;
  i = (int32_t)(x - y);
  if(i < 0) i += 1000000L;
  return i;
}

#ifndef __SPU__

void
cs_create_timer(struct sigevent *evp, timer_t *id,
	struct sigaction *act, void (*handler)(int), int signum) {
  int rc;

  evp->sigev_notify = SIGEV_SIGNAL;
  evp->sigev_signo = signum;
  rc = timer_create(CLOCK_REALTIME, evp, id);
  if(rc != 0) {
    perror("timer_create");
    GG_FATAL("Cannot create timer, rc=%d.\n", rc);
  }
  act->sa_handler = handler;
  act->sa_flags = 0;
  sigemptyset(&act->sa_mask);
  rc = sigaction(signum, act, NULL);
  if(rc != 0) {
    perror("sigaction");
    GG_FATAL("Cannot register signal handler, rc=%d.\n", rc);
  }
  return;
}

void
cs_delete_timer(timer_t id, struct sigaction *act) {
  if(timer_delete(id) != 0) {
    perror("timer_delete");
  }
  act->sa_handler = SIG_DFL;
  act->sa_flags = 0;
  sigemptyset(&act->sa_mask);
  if(sigaction(SIGALRM, act, NULL) != 0) {
    perror("Releasing sigaction");
  }
  return;
}

void
cs_set_timer(ggTime t, struct itimerspec *value, timer_t id) {
  int rc;
  value->it_value.tv_sec = floorf(t);
  value->it_value.tv_nsec = (t - floorf(t)) * 1e9;
  value->it_interval.tv_sec = 0;
  value->it_interval.tv_nsec = 0;
  rc = timer_settime(id, 0, value, NULL);
  if(rc != 0) {
    perror("timer_settime");
    GG_FATAL("Cannot set timer, rc=%d.\n", rc);
  }
  return;
}

#endif
