/***************************************************************
 * main.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <time.h>
#include <sys/time.h>
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <strings.h>
#include <sys/times.h>
#include "global.h"
#include "gggo.h"
#include "queue.h"

extern ggMove	gg_do_genmove(ggColor c);
extern void	gg_init_ren(void);
extern void	gg_init_playout(void);
extern int	gg_finish_playout(void);
extern void	gg_init_gen_rand(uint32_t seed);
extern void	gg_init_stack(void);
extern void	gg_init_block(void);
extern int	gg_count_final_score(ggColor c, int *n_dame);
#if USE_HASH
extern void	gg_create_hash(void);
extern void	gg_report_hash(FILE *);
#endif

void	gg_init_timers(void);
void	gg_push_clock_button(ggColor color, ggMove move);
ggTime	gg_colored_left(ggColor color);

#if PRINT_STATISTICS
FILE	*gg_report;
#endif

#if USE_ORACLE
FILE	*gg_oracle_from, *gg_oracle_to;
char	gg_oracle[GG_ORACLE_LINESIZE];
#endif

static const char *gg_program_name = "Fudo Go";
static const char *gg_version = "3.0";

/****************************************
 *	Common global variables		*
 ****************************************/
/* Game management variables */
int		gg_board_size = GG_MAX_BOARD;
float		gg_komi = 7.5;
int		gg_allow_suicide = false;	/* not used in mc() */
int		gg_resign_allowed = false;	/* not used */
int		gg_capture_all_dead = false;	/* not used */
int		gg_never_resign = false;
int		gg_never_pass = false;
int		gg_use_joseki = false;
ggScoring	gg_scoring = GG_AREA;
uint16_t	gg_main_time = 60 * 30;
uint8_t		gg_byoyomi_time = 0;		/* not used */
uint8_t		gg_byoyomi_stones = 0;		/* not used */

/* Various parameters */
ggRandomSeed	gg_random_seed = 0x4f321c57;
float		gg_fpu_initial = 1.10;
float		gg_resign_threshold = 0.1;
float		gg_ucb_factor = 0.45;
int		gg_n_internal = 3;
float		gg_k_rave = 2000;
uint32_t	gg_n_playout = GG_MAX_PLAYOUT;
int		gg_min_visits = 0;		/* Zero means every visits */
uint32_t	gg_generation = 0;
int		gg_eager_update = false;
int		gg_use_lan = false;
int		gg_fixed_window = 0;
float		gg_log_window = 0;
float		gg_mod_fpu = 0.99;
#if USE_LADDER_MODE
int		gg_use_ladder_mode = false;
#endif
#if USE_OFFSET
float		gg_offset = 0;
#endif
#if USE_EX_DELAY
int		gg_n_delay = 0;
#endif
ggTime		gg_time_per_move = 0;

bool		gg_in_joseki = true;

/* Game positions */
uint16_t	gg_bw_captured[GG_N_COLOR];
uint16_t	gg_movenum = 0;
ggKoInfo	gg_board_ko_info;
ggPoint		gg_previous_move = GG_NO_MOVE;
ggPoint		gg_previous_move2 = GG_NO_MOVE;
uint8_t		gg_n_pass = 0;
ggBool		gg_gameend = false;
ggMark		gg_mark_num = ~0;
ggMark		gg_mark_ren_num = ~0;
ggRenID		gg_ren_free = GG_REN_ROOT;
ggRenID		gg_max_ren_free = GG_REN_ROOT;

const ggDirection	gg_direction[8] = {
  GG_N_NORTH, GG_N_NORTH + GG_N_EAST, GG_N_EAST, GG_N_EAST + GG_N_SOUTH,
  GG_N_SOUTH, GG_N_SOUTH + GG_N_WEST, GG_N_WEST, GG_N_WEST + GG_N_NORTH };
const ggDirection	g2_direction[2][8] =
  {{ 0, 1, 1, 1, 0, -1, -1, -1},	/* x */
   { 1, 1, 0, -1, -1, -1, 0, 1 }};	/* y */


ggBoard		gg_board[GG_BOARDSIZE];
#if USE_EMPTY_LIST
ggPoint		gg_empty_list[GG_BOARDSIZE];
#endif
ggRen		gg_ren[GG_MAX_REN];
/************************
 *	Timers		*
 ************************/
static double	gg_time = 0;

static struct {
  ggTime 	cputime, cputime0;
  ggTime	elapsed, elapsed0;
  uint16_t	n_move;
} gg_colored_clock[GG_N_COLOR], *gg_current_clock = NULL;

static ggTime
gg_cputime(void) {
  clock_t rc;
  ggTime t;
  struct tms buf;

if(0) gprintf("--> %s\n", __func__);
  rc = times(&buf);
  if(rc != (clock_t)-1) {
    t = buf.tms_utime + buf.tms_stime + buf.tms_cutime + buf.tms_cstime;
    t /= sysconf(_SC_CLK_TCK);
  } else {
    perror(__func__);
    t = 0;
  }
if(0) gprintf("<-- %s %.1f\n", __func__, t);
  return t;
}

static double
gg_timeofday(void) {
  double t;
  int rc;
  struct timeval tv;

  rc = gettimeofday(&tv, NULL);
  if(rc == 0) {
    t = (double)tv.tv_sec + 1e-6 * (double)tv.tv_usec;
  } else {
    perror(__func__);
    t = 0;
  }
  return t;
}

static ggTime
gg_elapsedtime(void) {
  ggTime t;
if(0) gprintf("--> %s\n", __func__);
  t = gg_timeofday() - gg_time;
if(0) gprintf("<-- %s %.1f\n", __func__, t);
  return t;
}

void
gg_init_timers(void) {
  gg_time = gg_timeofday();
  gg_current_clock = NULL;
  gg_colored_clock[GG_C2X(GG_BLACK)].cputime = 0;
  gg_colored_clock[GG_C2X(GG_BLACK)].elapsed = 0;
  gg_colored_clock[GG_C2X(GG_BLACK)].n_move = 0;
  gg_colored_clock[GG_C2X(GG_BLACK)].elapsed0 = gg_elapsedtime();
  gg_colored_clock[GG_C2X(GG_BLACK)].cputime0 = gg_cputime();
  gg_colored_clock[GG_C2X(GG_WHITE)] = gg_colored_clock[GG_C2X(GG_BLACK)];
  return;
}

static ggTime
gg_colored_elapsed(ggColor c) {
  return (gg_current_clock == NULL) ? 0 : gg_colored_clock[GG_C2X(c)].elapsed;
}

ggTime
gg_colored_left(ggColor color) {
  return gg_main_time - gg_colored_elapsed(color);
}

/* TODO */
static void
gg_update_time_left(ggColor color, int time, int stones) {
#if 0
  ggTime delta;
  UNUSED(stones);	/* FIXME */
  delta = gg_colored_clock[GG_C2X(color)].elapsed -
		((ggTime)gg_main_time - (ggTime)time);
  gg_colored_clock[GG_C2X(color)].elapsed0 += delta;
#else
  UNUSED(stones);	/* FIXME */
  gg_colored_clock[GG_C2X(color)].elapsed = (ggTime)gg_main_time - (ggTime)time;
#endif
  return;
}

void
gg_push_clock_button(ggColor color, ggMove move) {
  ggTime t, f;
  uint32_t u;

if(0) gprintf("--> %s %C\n", __func__, color);
  if(gg_current_clock == NULL) {
    gg_current_clock = &gg_colored_clock[GG_C2X(color)];
  }
  t = gg_cputime() - gg_current_clock->cputime0;
  if(t < -0.01) {
    u = 0x80000000u;
    f = u;
    f /= (ggTime)CLOCKS_PER_SEC;
    t += f;
  }
  gg_current_clock->cputime += t;
  gg_current_clock->elapsed += gg_elapsedtime() - gg_current_clock->elapsed0;
  gg_current_clock->n_move += 1;
if(8) {
  gprintf(
    "Clock: color=%C move=%1m cpu=%.1fs elapsed=%.1fs left=%.1fs moves=%d\n",
    color, move, gg_current_clock->cputime, gg_current_clock->elapsed,
    gg_main_time - gg_current_clock->elapsed, gg_current_clock->n_move); 
  fflush(stderr);
}
#if PRINT_STATISTICS
  gfprintf(gg_report,
    "Clock: color=%C move=%1m cpu=%.2fs elapsed=%.2fs left=%.2fs moves=%d\n",
    color, move, gg_current_clock->cputime, gg_current_clock->elapsed,
    gg_main_time - gg_current_clock->elapsed, gg_movenum); 
    fflush(gg_report);
#endif
  gg_current_clock = &gg_colored_clock[GG_C2X(gg_opposit_color(color))];
  gg_current_clock->elapsed0 = gg_elapsedtime();
  gg_current_clock->cputime0 = gg_cputime();
if(0) gprintf("<-- %s\n", __func__);
  return;
}

static ggRandomSeed
gg_set_random_seed(ggRandomSeed i) {
  ggRandomSeed j;
  j = gg_random_seed;
  gg_random_seed = i;
  return j;
}

static void
gg_show_score(void) {
  ggScoreValue v;
if(0) gprintf("--> %s\n", __func__);
  v = gg_count_final_score(GG_WHITE, NULL) + gg_komi;
  printf("TEXT Final score: %.1f\n\n", v);
if(0) gprintf("<-- %s\n", __func__);
  return;
}

static void
gg_reset_engine(void) {
  gg_in_joseki = true;
#if USE_OFFSET
  gg_offset = 0;
#endif
  return;
}

/****************************************
 *	Printing functions		*
 ****************************************/
#define BLACK	GG_BLACK
#define WHITE	GG_WHITE
#define EMPTY	GG_EMPTY
#define EXTER	GG_EXTER
#define PASS_MOVE	GG_PASS_MOVE
#define RESIGN_MOVE	GG_RESIGN_MOVE
#define POS	GG_POS
#define X	GG_X
#define Y	GG_Y
#define is_not_ob	gg_is_not_ob
#define is_onboard2	gg_is_onboard2
#define x2c	gg_x2c
#include "printutils.c"

/****************************************
 *		GTP Handler		*
 ****************************************/
#define GTP_BUFLEN	1023
#define GTP_BUFSIZE	(GTP_BUFLEN + 1)

enum {
  GTP_QUIT = -1,
  GTP_OK,
  GTP_FATAL,
};

enum {
  GTP_SUCCESS = 0,
  GTP_FAILURE,
};

FILE		*gtp_infile, *gtp_outfile, *gtp_dumpfile;
static int	gtp_current_id;

/*****************************************************/

extern inline void gtp_putc(const char c);
extern inline void gtp_puts(const char *s);
#if HAS_FLOCK
extern inline void
gtp_putc(const char c) { putc_unlocked(c, gtp_outfile); }
extern inline void
gtp_puts(const char *s) { fputs_unlocked(s, gtp_outfile); }
#else
extern inline void
gtp_putc(const char c) { putc(c, gtp_outfile); }
extern inline void
gtp_puts(const char *s) { fputs(s, gtp_outfile); }
#endif

static void
gtp_printf(const char *format, ...) {
  va_list ap;
  va_start(ap, format);
  vfprintf(gtp_outfile, format, ap);
  va_end(ap);
  return;
}

static void
gtp_start_response(const int status) {
  gtp_putc(status == GTP_SUCCESS ? '=' : '?');
  if(gtp_current_id >= 0)
    gtp_printf("%d", gtp_current_id);
  gtp_putc(' ');
  return;
}

static void
gtp_finish_response(void) {
  gtp_putc('\r');
  gtp_putc('\n');
  gtp_putc('\r');
  gtp_putc('\n');
  return;
}

static void
gtp_panic(void) {
  gtp_puts("! PANIC");
  gtp_finish_response();
  return;
}

static int
gtp_reply(const int status, const char *format, ...) {
  va_list ap;
  gtp_start_response(status);
  va_start(ap, format);
  vfprintf(gtp_outfile, format, ap);
  va_end(ap);
  gtp_finish_response();
  return GTP_OK;
}

#define gtp_reply_success(format, ...) \
  gtp_reply(GTP_SUCCESS, format, ## __VA_ARGS__)

#define gtp_reply_failure(format, ...) \
  gtp_reply(GTP_FAILURE, format, ## __VA_ARGS__)

#define gtp_reply_success_null() \
  gtp_reply(GTP_SUCCESS, "")

static int
gtp_decode_color(const char *s, int *color) {
  char color_name[6 + 1];
  int n;

  if(sscanf(&s[0], "%" NAME2STR(6) "s%n", color_name, &n) != 1) {
    n = 0;
  } else {
    if(strcasecmp(color_name, "b") == 0 ||
       strcasecmp(color_name, "black") == 0) {
      *color = GG_BLACK;
    } else if(strcasecmp(color_name, "w") == 0 ||
	    strcasecmp(color_name, "white") == 0) {
      *color = GG_WHITE;
    } else {
      n = 0;
    }
  }
  return n;
}

static int
gtp_decode_coord(const char *s, int *i0, int *j0) {
  char c;
  int i, j;
  int n;

  if(sscanf(s, " %c%d%n", &c, &j, &n) != 2) {
    n = 0;
  } else {
    i = c;
    i = tolower(i);
    if(i == 'i') {
      n = 0;
    } else {
      i = i - 'a' - (i > 'i' ? 1 : 0);
      j -= 1;
      if(0 <= i && i < gg_board_size && 0 <= j && j < gg_board_size) {
	*i0 = i;
	*j0 = j;
      } else {
	n = 0;
      }
    }
  }
  return n;
}

extern ggMove	gg_str2cord(const char *s, const char **s0);
ggMove
gg_str2cord(const char *s, const char **s0) {
  int i, j, n;
  ggMove move;

if(0) gprintf("--> %s \"%s\"\n", __func__, s);
  n = gtp_decode_coord(s, &i, &j);
  if(n > 0) {
    move = GG_POS(i, j);
    *s0 = s + n;
  } else {
    move = 0;
  }
if(0) gprintf("<-- %s move=%1m\n", __func__, move);
  return move;
}

static void
gtp_print_vertex(int i, int j) {
  if(i == -1 && j == -1) {
    gtp_puts("pass");
  } else if(!(0 <= i && i < gg_board_size && 0 <= j && j < gg_board_size)) {
    gtp_puts("????");
  } else {
    gtp_printf("%c%d", gg_x2c(i), j + 1);
  }
  return;
}

/* Preprocessing.
 */
static void
gtp_preprocess(char *p) {
  char c, *q;
  q = p;
  while(true) {
    c = *p++;
    if(c == '\0' || c == '\n' || c == '\r' || c == '#')	break;
    if(c < ' ' || c == 0x7f)	continue;
    if(c == '\t')		c = ' ';
    *q++ = c;
  }
  *q = '\0';
  return;
}

static char *
gtp_sequence_id(char *p, int *id0) {
  int n;

if(0) fprintf(stderr, "--> %s(\"%s\")\n", __func__, p);
  if(*p == '=')	p++;
  if(sscanf(p, "%d%n", id0, &n) == 1) {
    p += n;
  } else {
    *id0 = -1;
  }
if(0) fprintf(stderr, "<-- %s id=%d p=\"%s\" \n", __func__, *id0, p);
  return p;
}

#if USE_ORACLE
//static int	gtp_oracle_id = 0;

char *
gg_get_oracle(char *p) {
  int id;

if(0) fprintf(stderr, "--> %s(%p)\n", __func__, p);
  fgets(p, GG_ORACLE_LINESIZE, gg_oracle_from);
  gtp_preprocess(p);
  p = gtp_sequence_id(p, &id);
if(0) fprintf(stderr, "Got from oracle \"%s\" -- %s.\n", p, __func__);
  getc(gg_oracle_from);		/* Discard '\n' */
if(0) fprintf(stderr, "<-- %s \"%s\"\n", __func__, p);
  return p;
}

void
gg_put_oracle(const char *format, ...) {
  va_list ap;
if(0) fprintf(stderr, "--> %s(%s, ...)\n", __func__, format);
#if 0
  gtp_oracle_id += 1;
  putc('=', gg_oracle_to);
  if(gtp_oracle_id >= 0)	fprintf(gg_oracle_to, "%d", gtp_oracle_id);
  putc(' ', gg_oracle_to);
#endif
  va_start(ap, format);
  vgprintf(gg_oracle_to, format, ap);
  va_end(ap);
  putc('\n', gg_oracle_to);
  putc('\n', gg_oracle_to);
  fflush(gg_oracle_to);
  return;
}
#endif

/*****************************************************/

typedef struct {
  const char	*name;
  int		(*func)(const char *s);
} gtpCommand;

#define GTP_COMMAND(func)	static int func(const char *s)
GTP_COMMAND(gtp_quit);
GTP_COMMAND(gtp_protocol_version);
GTP_COMMAND(gtp_name);
GTP_COMMAND(gtp_program_version);
GTP_COMMAND(gtp_set_boardsize);
GTP_COMMAND(gtp_clear_board);
GTP_COMMAND(gtp_set_komi);
GTP_COMMAND(gtp_get_komi);
GTP_COMMAND(gtp_playblack);
GTP_COMMAND(gtp_playwhite);
GTP_COMMAND(gtp_play);
GTP_COMMAND(gtp_genmove_black);
GTP_COMMAND(gtp_genmove_white);
GTP_COMMAND(gtp_genmove);
GTP_COMMAND(gtp_undo);
GTP_COMMAND(gtp_time_settings);
GTP_COMMAND(gtp_time_left);
GTP_COMMAND(gtp_cputime);
GTP_COMMAND(gtp_known_command);
GTP_COMMAND(gtp_list_commands);
GTP_COMMAND(gtp_final_score);
#if 0
GTP_COMMAND(gtp_final_status_list);
GTP_COMMAND(gtp_load_sgf);
GTP_COMMAND(gtp_query_boardsize);
GTP_COMMAND(gtp_fixed_handicap);
GTP_COMMAND(gtp_place_free_handicap);
GTP_COMMAND(gtp_set_free_handicap);
#endif
GTP_COMMAND(gtp_show_score);

/* List of supported gtp commands. */
static gtpCommand gtp_commands[] = {
  {"quit",			gtp_quit},
  {"protocol_version",		gtp_protocol_version},
  {"name",			gtp_name},
  {"version",			gtp_program_version},
  {"boardsize",			gtp_set_boardsize},
  {"clear_board",		gtp_clear_board},
  {"komi",			gtp_set_komi},
  {"get_komi",			gtp_get_komi},
  {"playblack",			gtp_playblack},
  {"playwhite",			gtp_playwhite},
  {"play",			gtp_play},
  {"genmove_black",		gtp_genmove_black},
  {"genmove_white",		gtp_genmove_white},
  {"genmove",			gtp_genmove},
  {"undo",			gtp_undo},
  {"time_settings",		gtp_time_settings},
  {"time_left",			gtp_time_left},
  {"cputime",			gtp_cputime},
  {"known_command",		gtp_known_command},
  {"list_commands",		gtp_list_commands},
  {"help",			gtp_list_commands},
  {"final_score",		gtp_final_score},
#if 0
  {"final_status_list",		gtp_final_status_list},
  {"load_sgf",			gtp_load_sgf},
  {"query_boardsize",		gtp_query_boardsize},
  {"fixed_handicap",		gtp_fixed_handicap},
  {"place_free_handicap",	gtp_place_free_handicap},
  {"set_free_handicap",		gtp_set_free_handicap},
#endif
  {"show_score",		gtp_show_score},
  {NULL,			NULL},
};

static int
gtp_quit(const char *s) {
  UNUSED(s);
#if USE_ORACLE
  gg_put_oracle("quit");
#endif
  gtp_reply_success_null();
  return GTP_QUIT;
}

static int
gtp_protocol_version(const char *s) {
  UNUSED(s);
  return gtp_reply_success("2");
}

static int
gtp_name(const char *s) {
  UNUSED(s);
  return gtp_reply_success(gg_program_name);
}

static int
gtp_program_version(const char *s) {
  UNUSED(s);
  return gtp_reply_success(gg_version);
}

static int
gtp_clear_board(const char *s) {
  UNUSED(s);
  gtp_reply_success_null();
  gg_clear_board();
  gg_reset_engine();
  gg_init_timers();
#if USE_ORACLE
  gg_put_oracle("clear_board");
  gg_get_oracle(&gg_oracle[0]);
#endif
if(0) { putc('\n', stderr); fflush(stderr); }
  return GTP_OK;
}

static int
gtp_set_boardsize(const char *s) {
  int boardsize;

  if(sscanf(s, "%d", &boardsize) < 1)
    return gtp_reply_failure("boardsize must be an integer");
  if(boardsize > GG_MAX_BOARD)
    return gtp_reply_failure("unacceptable size");
  gg_board_size = boardsize;
  gtp_reply_success_null();
  gg_clear_board();
  gg_reset_engine();
  gg_init_timers();
#if USE_ORACLE
  gg_put_oracle("boardsize %d", gg_board_size);
  gg_get_oracle(&gg_oracle[0]);
#endif
  return GTP_OK;
}

static int
gtp_set_komi(const char *s) {
  float gtp_komi;

  if(sscanf(s, "%f", &gtp_komi) < 1)
    return gtp_reply_failure("komi must be a number");
  gg_komi = gtp_komi;
  gtp_reply_success_null();
#if USE_ORACLE
  gg_put_oracle("komi %.1f", gg_komi);
  gg_get_oracle(&gg_oracle[0]);
#endif
  return GTP_OK;
}

static int
gtp_get_komi(const char *s) {
  UNUSED(s);
  return gtp_reply_success("%.1f", gg_komi);
}

static int
gtp_do_play(ggMove move, ggColor c) {
  gtp_reply_success_null();
  gg_play_move(move, c);
#if USE_ORACLE
  gg_put_oracle("play %C %1m", c, move);
  gg_get_oracle(&gg_oracle[0]);
#endif
  return GTP_OK;
}

static int
gtp_play_bw(const char *s, const int c) {
  int i, j;

  i = -1;
  j = -1;
  if(strncasecmp("pass", s, 4) != 0) {
    gtp_decode_coord(&s[0], &i, &j);
    if(gg_is_illegal(GG_POS(i, j), c))
      return gtp_reply_failure("illega move");
  }
  return gtp_do_play(GG_POS(i, j), c);
}

static int
gtp_playblack(const char *s) {
  return gtp_play_bw(s, GG_BLACK);
}

static int
gtp_playwhite(const char *s) {
  return gtp_play_bw(s, GG_WHITE);
}

static int
gtp_play(const char *s) {
  int c;
  int i, j;

  int n, n2;
  char str[6];

  n = gtp_decode_color(&s[0], &c);
  if(n == 0)
    return gtp_reply_failure("invalid color");

  n2 = gtp_decode_coord(&s[n], &i, &j);
  if(n2 == 0) {
    if(sscanf(&s[n], "%5s%n", str, &n2) == 1 && strcasecmp(str, "pass") == 0) {
      i = -1;
      j = -1;
    } else {
      return gtp_reply_failure("invalid coordinate");;
    }
  }
  if(gg_is_illegal(GG_POS(i, j), c))
    return gtp_reply_failure("illegal move");
  return gtp_do_play(GG_POS(i, j), c);
}

static void
gtp_do_genmove(ggColor color) {
  const ggMove move = gg_genmove(color);
  gtp_start_response(GTP_SUCCESS);
  if(move == GG_RESIGN_MOVE) {
    gtp_puts("resign");
    gtp_finish_response();
  } else {
    gtp_print_vertex(GG_X(move), GG_Y(move));
    gtp_finish_response();
    gg_play_move(move, color);
#if USE_ORACLE
    gg_put_oracle("play %C %1m", color, move);
    gg_get_oracle(&gg_oracle[0]);
#endif
  }
}

static int
gtp_genmove_black(const char *s) {
  UNUSED(s);
  gtp_do_genmove(GG_BLACK);
  return GTP_OK;
}

static int
gtp_genmove_white(const char *s) {
  UNUSED(s);
  gtp_do_genmove(GG_WHITE);
  return GTP_OK;
}

static int
gtp_genmove(const char *s) {
  int color;
  int n;

if(0) gprintf("--> %s\n", __func__);
  n = gtp_decode_color(s, &color);
  if(n < 1)	return gtp_reply_failure("invalid color");
  gtp_do_genmove(color);
if(0) gprintf("<-- %s\n", __func__);
  return GTP_OK;
}

static int
gtp_undo(const char *s) {
  UNUSED(s);
  if(gg_undo_move(1) == false) {
    return gtp_reply_failure("cannot undo");
  } else {
    gtp_reply_success_null();
#if USE_ORACLE
    gg_put_oracle("undo");
    gg_get_oracle(&gg_oracle[0]);
#endif
    return GTP_OK;
  }
}

static int
gtp_time_settings(const char *s) {
  int main_time, byoyomi_time, byoyomi_stones;

  if(sscanf(s, "%d %d %d", &main_time, &byoyomi_time, &byoyomi_stones) < 3)
    return gtp_reply_failure("time settings must be three integers");
  gtp_reply_success_null();
  gg_main_time = main_time;
  gg_byoyomi_time = byoyomi_time;
  gg_byoyomi_stones = byoyomi_stones;
#if USE_ORACLE
  gg_put_oracle("time_settings %d %d %d",
			  main_time, byoyomi_time, byoyomi_stones);
  gg_get_oracle(&gg_oracle[0]);
#endif
  return GTP_OK;
}

static int
gtp_time_left(const char *s) {
  int color;
  int n;
  int time, stones;

  n = gtp_decode_color(&s[0], &color);
  if(n < 1)
    return gtp_reply_failure("invalid color");
  n = sscanf(&s[n], "%d %d", &time, &stones);
  if(n < 2)
    return gtp_reply_failure("time and stones must be two integers");
  gtp_reply_success_null();
  gg_update_time_left(color, time, stones);
#if USE_ORACLE
  gg_put_oracle("time_left %C %d %d", color, time, stones);
  gg_get_oracle(&gg_oracle[0]);
#endif
  return GTP_OK;
}

static int
gtp_final_score(const char *s) {
  ggRandomSeed old_seed, seed;
  ggScoreValue final_score;

  seed = 0;
  sscanf(&s[0], "%d", &seed);
  old_seed = gg_set_random_seed(seed);
  final_score = gg_finish_and_score(seed);
  gg_set_random_seed(old_seed);

  gtp_start_response(GTP_SUCCESS);
  if(final_score > 0) {
    gtp_printf("W+%3.1f", final_score);
  } else if(final_score < 0) {
    gtp_printf("B+%3.1f", -final_score);
  } else {
    gtp_printf("0");
  }
  gtp_finish_response();
  return GTP_OK;
}

#if 0
#define GG_REN_STATUS_MAX	6

static const char *gg_status_names[] =
  {"alive", "dead", "seki", "white_territory", "black_territory", "dame"};
ggRenStatus gg_final_status[GG_MAX_BOARD][GG_MAX_BOARD];

static int
gtp_final_status_list(const char *s) {
  int n;
  ggRenStatus status;
  ggRandomSeed old_seed, seed;
  ggScoreValue final_score;
  bool first;
  char status_str[GTP_BUFSIZE];

  if(sscanf(&s[0], "%" NAME2STR(GTP_BUFLEN) "s %n", &status_str[0], &n) != 1)
    return gtp_reply_failure("missing status");

  status = 0;
  for(int i = 0; i < GG_REN_STATUS_MAX; i++) {
    if(strcasecmp(&status_str[0], status_names[i]) == 0)
      status = i + 1;		/* Assume ALIVE = 1, ... */
  }
  if(status == 0)	return gtp_reply_failure("invalid status");

  seed = 0;
  sscanf(&s[n], "%d", &seed);
  old_seed = gg_set_random_seed(seed);
  final_score = gg_finish_and_score(seed);
  gg_set_random_seed(old_seed);

  gtp_start_response(GTP_SUCCESS);
  first = true;
  for(int j = 0; j < board_size; j++) {
    for(int i = 0; i < board_size; i++) {
      if(gg_final_status[j][i] != status)	continue;
      if(!first) {
      	gtp_putc(' ');
      } else {
      	first = false;
      }
      gtp_print_vertex(i, j);
    }
  }
  gtp_finish_response();
  return GTP_OK;
}
#endif

static int
gtp_cputime(const char *s) {
  UNUSED(s);
  return gtp_reply_success("%.3f", gg_cputime());
}

static int
gtp_list_commands(const char *s) {
  UNUSED(s);
  gtp_start_response(GTP_SUCCESS);
  for(int i = 0; gtp_commands[i].name != NULL; i++) {
    gtp_puts(gtp_commands[i].name);
    gtp_putc('\n');
  }
  gtp_putc('\n');
  return GTP_OK;
}

static int
gtp_known_command(const char *s) {
  bool found;
  char command[GTP_BUFSIZE];

  found = true;
  if(sscanf(s, "%" NAME2STR(GTP_BUFLEN) "s", &command[0]) == 1) {
    for(int i = 0; gtp_commands[i].name != NULL; i++)
      if(strcasecmp(command, gtp_commands[i].name) == 0)	goto exit;
  }
  found = false;
exit:
  return gtp_reply_success(found ? "true" : "false");
}

static int
gtp_show_score(const char *s) {
  UNUSED(s);
  gtp_start_response(GTP_SUCCESS);
  gg_show_score();
  return GTP_OK;
}

/****************************************
 *		main loop		*
 ****************************************/
static void
gtp_mainloop(void) {
  char *p, line[GTP_BUFSIZE], command[GTP_BUFSIZE];
  int status;
  int i, n;

if(0) gprintf("--> %s()\n", __func__);
  status = GTP_OK;
  do {
    p = fgets(&line[0], GTP_BUFSIZE, gtp_infile);
    if(p == NULL)	break;
    if(gtp_dumpfile != NULL) {
      fputs(p, gtp_dumpfile);
      fflush(gtp_dumpfile);
    }

    gtp_preprocess(&line[0]);
    p = gtp_sequence_id(&line[0], &gtp_current_id);

    if(sscanf(p, " %" NAME2STR(GTP_BUFLEN) "s %n", &command[0], &n) > 0) {
      p += n;
      for(i = 0; ; i++) {
	if(gtp_commands[i].name == NULL) {
	  gtp_reply_failure("unknown command");
	  break;
	} else if(strcasecmp(command, gtp_commands[i].name) == 0) {
    if(0) fprintf(stderr, "--> %s %s\n", __func__, command);
	  status = (*gtp_commands[i].func)(p);
    if(0) fprintf(stderr, "<-- %s %s rc=%d\n", __func__, command, status);
	  break;
	}
      }
      if(status == GTP_FATAL)	gtp_panic();
    }
  } while(status == GTP_OK);
if(0) gprintf("<-- %s\n", __func__);
  return;
}

/********************************************************
 *		Thread and Communication		*
 ********************************************************/

typedef struct {
  ggQueue	*in, *out;
} ggDoubleQueue;

typedef uint32_t	ggMessageInfo;	/* Lower 24 bits are effective. */

typedef enum {
  TypeAsk	= 1 << 4,
  TypeInterrupt	= 2 << 4,
} ggMessageType;

typedef enum {
  KindMove = 1,
  KindQuit,
} ggMessageKind;

extern inline ggMessage gg_async_pop(ggQueue *queue);
extern inline ggMessage
gg_async_pop(ggQueue *queue) {
  return gg_pop(queue, NULL);
}

extern inline int gg_async_push(ggQueue *queue, ggMessage message);
extern inline int
gg_async_push(ggQueue *queue, ggMessage message) {
  return gg_push(message, queue);
}

extern inline ggMessage
gg_pack_message(ggMessageType type, ggMessageKind kind, ggMessageInfo info);
extern inline ggMessage
gg_pack_message(ggMessageType type, ggMessageKind kind,
      ggMessageInfo info) {
  return (((ggMessage)(type | kind) << (32 - 8)) | info);
}

extern inline int
gg_unpack_message(ggMessage message, ggMessageType *type, ggMessageKind *kind);
extern inline int
gg_unpack_message(ggMessage message, ggMessageType *type,
      ggMessageKind *kind) {
  ggMessageInfo info = message & 0xffffff;
  message >>= (32 - 8);
  *type = message & 0xf0;
  *kind = message & 0x0f;
  return info;
}

/******************************************************
 *		Thread Interface for genmove	      *
 ******************************************************/
static ggQueue	gg_genmove_queue_in, gg_genmove_queue_out;
static ggDoubleQueue gg_genmove_queue;

extern inline ggMessage gg_pack_move(ggMove move, ggColor color);
extern inline ggMessage
gg_pack_move(ggMove move, ggColor color) {
  return gg_pack_message(TypeAsk, KindMove, (move << 4 || color));
}

static void
gg_genmove_thread(ggDoubleQueue *genmove_queue0) {
  ggMessage message;
  ggMessageType type;
  ggMessageKind kind;
  ggMessageInfo info;
  ggColor color;
  ggMove move;
  int rc;
  bool exit;

if(0) gprintf("--> %s: %p\n", __func__, genmove_queue0);
  assert(genmove_queue0 != NULL);
  assert(genmove_queue0->in != NULL);
  assert(genmove_queue0->out != NULL);

  exit = false;
  do {
if(0) gprintf("--- %s: waiting in=%p\n", __func__, genmove_queue0->in);
    message = gg_pop(genmove_queue0->in, &rc);
    if(rc != 0)		break;
    info = gg_unpack_message(message, &type, &kind);
if(0) gprintf("--- %s: received type=%#2x kind=%#1x info=%#3x\n", __func__,
			  type, kind, info);
    switch(type | kind) {
      case (TypeAsk | KindMove):
	color = info & 0xff;
	move = gg_do_genmove(color);
if(0) gprintf("--- %s: sending... move=%1m color=%C port=%p\n", __func__,
			  move, color, genmove_queue0->out);
	rc = gg_push(gg_pack_message(TypeAsk, KindMove, move),
					    genmove_queue0->out);
	if(rc != 0)	exit = true;
if(0) gprintf("--- %s: done.\n", __func__);
      break;

      case (TypeInterrupt | KindQuit):
if(0) fprintf(stderr, "--- %s: Quiting...\n", __func__);
	exit = true;
      break;

      default:
	fprintf(stderr, "Abort! %s message=%x\n", __func__, message);
	abort();
      break;
    }
  } while(exit == false);
if(0) fprintf(stderr, "<-- %s\n", __func__);
  return;
}

static pthread_t
gg_create_genmove_thread(void (*start)(ggDoubleQueue *), void *arg) {
  int rc;
  pthread_t tid;
  pthread_attr_t attribute;

  pthread_attr_init(&attribute);
  pthread_attr_setdetachstate(&attribute, PTHREAD_CREATE_DETACHED);
  rc = pthread_create(&tid, &attribute, (void *(*)(void *))start, arg);
  if(rc != 0) {
    perror(__func__);
    abort();
  }
  return tid;
}

ggMove
gg_genmove(ggColor color) {
  ggMove move;
  ggMessage message;
  ggMessageType type;
  ggMessageKind kind;

if(0) gprintf("--> %s %C\n", __func__, color);
  assert(gg_genmove_queue.in != NULL);
  assert(gg_genmove_queue.out != NULL);

if(0) gprintf("--- %s queue=%p in=%p out=%p\n", __func__,
	  &gg_genmove_queue, gg_genmove_queue.in, gg_genmove_queue.out);

  gg_push(gg_pack_message(TypeAsk, KindMove, color), gg_genmove_queue.in);
if(0) gprintf("---> %s sent %C\n", __func__, color);

  message = gg_pop(gg_genmove_queue.out, NULL);
  move = gg_unpack_message(message, &type, &kind);
if(0) gprintf("<--- %s received move=%1m\n", __func__, move);

if(0) gprintf("<-- %s %1m\n", __func__, move);
  return move;
}

/************************************************
 *		Main Program			*
 ************************************************/
static void
gg_print_help(const char *command, FILE *stream) {
  fprintf(stream, "Usage: %s [options]\n", command);
  fprintf(stream, " -v, --version:\tprint version.\n");
  fprintf(stream, " -h, --help:\tprint this help.\n");
  fprintf(stream, "     --boardsize=<boardsize>:\tset boardsize.\n");
  fprintf(stream, "     --komi=<komi>:\tset komi.\n");
  fprintf(stream, "     --time=<second>:\tset time for the game.\n");
  fprintf(stream, "     --time-per-move=<second>:\tset time for a move.\n");
  fprintf(stream, "     --playout=<playout>:\tset memory size being used.\n");
  fprintf(stream, "     --seed=<seed>:\tset random seed.\n");
  fprintf(stream, "     --fpu=<fpu>:\tset initial value of fpu.\n");
  fprintf(stream, "     --never-resign:\tforce never resign.\n");
  /* TODO */
  return;
}

#if USE_DYNAMIC_ALLOCATE
extern char	*gg_init_uct(uint32_t n);
extern char	*gg_init_position(uint32_t n);
#endif

int
main(int argc, char *argv[]) {
  pthread_t	genmove_tid;
  int i, rc;
  time_t t;
#if GENERATE_PATTERN
extern void	mc_print_pattern(void);
#endif

  /* Parse args. */
#ifndef __CYGWIN__
extern char	*optarg;
extern int	optind, opterr, optopt;
#endif
  int	c, index;
  char 	*endptr;
  ggColor color = GG_BLACK;
  const char *rules;
#if USE_ORACLE
  int pipe_a[2], pipe_b[2];
  pid_t metamachine = 0;
  char mm_komi[64], mm_boardsize[64], mm_level[] = "10";
#endif

static struct option long_options[] = {
  { "help",		no_argument,		NULL, 'h' },
  { "version",		no_argument,		NULL, 'v' },
  { "boardsize",	required_argument,	NULL, 1 },
  { "komi",		required_argument,	NULL, 2 },
  { "mode",		required_argument,	NULL, 3 },
  { "level",		required_argument,	NULL, 4 },
  { "color",		required_argument,	NULL, 5 },
  { "infile",		required_argument,	NULL, 6 },
  { "outfile",		required_argument,	NULL, 7 },
  { "logfile",		required_argument,	NULL, 8 },
  { "time",		required_argument,	NULL, 9 },
  { "byo-time",		required_argument,	NULL, 10 },
  { "byo-period",	required_argument,	NULL, 11 },
  { "rules",		required_argument,	NULL, 12 },
  { "chinese-rules",	no_argument,		NULL, 13 },
  { "japanese-rules",	no_argument,		NULL, 14 },
  { "seed",		required_argument,	NULL, 15 },
  { "playout",		required_argument,	NULL, 16 },
  { "fpu",		required_argument,	NULL, 17 },
  { "resign",		required_argument,	NULL, 18 },
  { "ucb-factor",	required_argument,	NULL, 20 },
  { "cp",		required_argument,	NULL, 20 },
  { "internal",		required_argument,	NULL, 21 },
  { "k_rave",		required_argument,	NULL, 22 },
  { "min-visits",	required_argument,	NULL, 23 },
  { "fixed-window",	required_argument,	NULL, 24 },
  { "log-window",	required_argument,	NULL, 25 },
  { "mod-fpu",		required_argument,	NULL, 26 },
  { "time-per-move",	required_argument,	NULL, 27 },
  { "generate-pattern",	no_argument,		NULL, 31 },
#if USE_EX_DELAY
  { "experimental-delay", required_argument,	NULL, 32 },
#endif
  { "allow-suicide",	no_argument,	&gg_allow_suicide, true },
  { "resign-allowed",	no_argument,	&gg_resign_allowed, true },
  { "capture-all-dead",	no_argument,	&gg_capture_all_dead, true },
  { "never-resign",	no_argument,	&gg_never_resign, true },
  { "never-pass",	no_argument,	&gg_never_pass, true },
  { "use-joseki",	no_argument,	&gg_use_joseki, true },
  { "eager-update",	no_argument,	&gg_eager_update, true },
#if USE_LADDER_MODE
  { "use-ladder-mode",	no_argument,	&gg_use_ladder_mode, true },
#endif
  { "use-lan",		no_argument,	&gg_use_lan, true },
  { NULL, 0, NULL, 0 } };

  gg_allow_suicide = false;
  gg_resign_allowed = false;
  gg_capture_all_dead = false;
#if USE_LADDER_MODE
  gg_use_ladder_mode = false;
#endif
#if 1
{
  struct timeval tv;
  int rc = gettimeofday(&tv, NULL);
  gg_random_seed =
    (ggRandomSeed)(rc == 0 ? tv.tv_sec * 1000000 + tv.tv_usec : 0x4f321c57);
}
#else
  gg_random_seed = 0x4f321c57;
#endif
  gg_n_internal = 3;		/* TODO number of cores minus 1 */

  rc = 1;
  do {
    c = getopt_long(argc, argv, "hvl:o:", long_options, &index);
    if(c < 0)	break;
    switch(c) {
      default:
	gg_print_help(argv[0], stderr);
	rc = 1;
	goto exit;
      break;
      case 'h':	/* print help. */
	gg_print_help(argv[0], stdout);
	goto exit0;
      break;
      case 'v':	/* print version. */
	printf("%s v%s\n", gg_program_name, gg_version);
	goto exit0;
      break;
      case 'l':	/* load sgf file. */
	/* TODO */
      break;
      case 'o':	/* output sgf file. */
	/* TODO */
      break;

      case 0:
      break;

      case 1: {	/* boardsize */
	gg_board_size = strtoul(optarg, &endptr, 10);
	if(*optarg == '\0' || *endptr != '\0' || gg_board_size > GG_MAX_BOARD) {
	  puts("Error: boardsize > " NAME2STR(GG_MAX_BOARD) ".\n");
	  goto exit;
	}
      }
      break;

      case 2: {	/* komi */
	float komi;
	komi = strtod(optarg, &endptr);
	if(*optarg == '\0' || *endptr != '\0') {
	  fprintf(stderr, "Error: Komi must be a number\n");
	  goto exit;
	} else {
	  gg_komi = komi;
	}
      }
      break;

      case 5:	/* color	CAUTION: This has no effect now. */
	if(strcasecmp(optarg, "black") == 0) {
	  color = GG_BLACK;
	} else if(strcasecmp(optarg, "white") == 0) {
	  color = GG_WHITE;
	} else {
	  printf("Error: Unknown color \"%s\".\n", optarg);
	  goto exit;
	}
      break;

      case 9:	/* clock */
	gg_main_time = strtoul(optarg, &endptr, 10);
      break;

      case 3:	/* mode */
      case 4:	/* level */
      case 6:	/* infile */
      case 7:	/* outfile */
      case 8:	/* logfile */
      case 10:	/* byo-time */
      case 11:	/* byo-period */
	puts("Sorry, not implemented yet.\n");
	goto exit;
      break;

      case 14:
	rules = "japanese";
	goto set_rules;
      case 13:
	rules = "chinese";
	goto set_rules;
      case 12:
	rules = optarg;
set_rules:
	if(strcasecmp(rules, "japanese") == 0) {
	  gg_allow_suicide = false;
	  gg_scoring = GG_TERRITORY;
	} else if(strcasecmp(rules, "chinese") == 0) {
	  gg_allow_suicide = false;
	  gg_scoring = GG_AREA;
	} else {
	  printf("Error: Unknown rules \"%s\".\n", rules);
	  goto exit;
	}
      break;

      case 15:		/* seed */
	gg_random_seed = strtoul(optarg, &endptr, 10);
      break;

      case 16: 		/* playout */
	gg_n_playout = atoi(optarg);
      break;
      case 17:		/* fpu=<float> */
	gg_fpu_initial = strtof(optarg, &endptr);
      break;
      case 18:		/* resign=<float> */
	gg_resign_threshold = strtof(optarg, &endptr);
      break;
      case 20:		/* ucb-factor=<float> */
	gg_ucb_factor = strtof(optarg, &endptr);
      break;
      case 21:		/* number of internal mc servers */
	gg_n_internal = atoi(optarg);
      break;
      case 22:		/* The constant K in RAVE */
	gg_k_rave = strtof(optarg, &endptr);
      break;
      case 23:		/* Minimum visits to create a node */
	gg_min_visits = atoi(optarg);
      break;
      case 24:		/* Experimental. Ignore late results */
	gg_fixed_window = atoi(optarg);
      break;
      case 25:
	gg_log_window = strtof(optarg, &endptr);
      break;
      case 26:
	gg_mod_fpu = strtof(optarg, &endptr);
      break;
      case 27:
	gg_time_per_move = strtof(optarg, &endptr);
      break;
#if GENERATE_PATTERN
      case 31:		/* Print interest pattern table */
	mc_print_pattern();
	return 0;
      break;
#endif
#if USE_EX_DELAY
      case 32:		/* Experimental test how delay works */
	gg_n_delay = atoi(optarg);
      break;
#endif
    }
  } while(true);


  if(gg_n_playout > GG_MAX_PLAYOUT) {
    puts("Error: Total playouts exceeds " NAME2STR(GG_MAX_PLAYOUT) ".\n");
    goto exit;
  }

#if USE_DYNAMIC_ALLOCATE
{ char *s;
  s = gg_init_uct((gg_n_playout + 1 +
	    gg_n_internal + (USE_EX_DELAY ? MAX_EX_DELAY : 0)) * 6 / 5);
  if(s != NULL) {
    printf("Error: %s.\n", s);
    goto exit;
  }
  s = gg_init_position((gg_n_playout + 1 +
			2 * gg_board_size * gg_board_size) * 6 / 5);
  if(s != NULL) {
    printf("Error: %s.\n", s);
    goto exit;
  }
}
#endif

  if(gg_fixed_window == 0 && gg_log_window == 0) {
    gg_fixed_window = INT_MAX;
  }

  t = time(NULL);
  fprintf(stderr, "Game start at %s", asctime(localtime(&t)));
  gprintf("playout=%d fpu=%.2f mod-fpu=%.2f seed=%u"
	  " Cp=%.3f internal=%d use-lan=%b eager-update=%b",
      gg_n_playout, gg_fpu_initial, gg_mod_fpu, gg_random_seed,
      gg_ucb_factor, gg_n_internal, gg_use_lan, gg_eager_update);
  if(gg_fixed_window != 0) {
    fprintf(stderr, " fixed-window=%u", gg_fixed_window);
  } else if(gg_log_window != 0) {
    fprintf(stderr, " log-window=%.1f", gg_log_window);
  }
  gprintf(" time/move=%.2fs", gg_time_per_move);
#if USE_LADDER_MODE
  gprintf(" ladder-mode=%b", gg_use_ladder_mode);
#endif
#if USE_EX_DELAY
  fprintf(stderr, " ex-delay=%d", gg_n_delay);
#endif
  putc('\n', stderr);

  /* Initilize etc. */
if(0) fprintf(stderr, "Initilize start... -- %s\n", __func__);
  gg_init_gen_rand((uint32_t)gg_random_seed);
  gg_init_stack();
  gg_clear_board();
  gg_init_block();
  gg_init_playout();
#if USE_HASH
  gg_create_hash();
#endif

  /* Setup thread */
if(0) fprintf(stderr, "Setup start... -- %s\n", __func__);
  gg_init_queue(&gg_genmove_queue_in);
  gg_init_queue(&gg_genmove_queue_out);
  gg_genmove_queue.in = &gg_genmove_queue_in;
  gg_genmove_queue.out = &gg_genmove_queue_out;
  genmove_tid = gg_create_genmove_thread(&gg_genmove_thread, &gg_genmove_queue);
if(0) fprintf(stderr, "Created genmove_thread %#lx\n", (long unsigned int)genmove_tid);
  gg_generation = 0;
  gg_reset_engine();

#if USE_ORACLE
  /* Invoke GNU Go */
  rc = pipe(pipe_a);
  if(rc != 0) {
    perror(__func__);
    GG_FATAL("Can't create pipe_a rc=%d\n", rc);
  }
  rc = pipe(pipe_b);
  if(rc != 0) {
    perror(__func__);
    GG_FATAL("Can't create pipe_b rc=%d\n", rc);
  }
  metamachine = fork();
  if(metamachine == 0) {
    rc = dup2(pipe_a[0], 0);
    if(rc == -1) {
      perror(__func__);
      GG_FATAL("Can't duplicate pipe_a rc=%d\n", rc);
    }
    rc = dup2(pipe_b[1], 1);
    if(rc == -1) {
      perror(__func__);
      GG_FATAL("Can't duplicate pipe_b rc=%d\n", rc);
    }
    sprintf(&mm_komi[0], "%.1f ", gg_komi);
    sprintf(&mm_boardsize[0], "%d", gg_board_size);
    execlp("gnugo", "gnugo", "--mode", "gtp", "--quiet", "--play-out-aftermath",
      "--chinese-rules", "--capture-all-dead", "--level", &mm_level[0],
      "--max-level", &mm_level[0], "--min-level", &mm_level[0],
      "--komi", &mm_komi[0], "--boardsize", &mm_boardsize[0],
      "--color", (color == GG_WHITE ? "white" : "black"), (char *)NULL);
    perror(__func__);
    GG_FATAL("Can't execlp GNU Go\n");
  } else if(metamachine == -1) {
    perror(__func__);
    GG_FATAL("Can't fork\n");
  }
  gg_oracle_to = fdopen(pipe_a[1], "w");
  gg_oracle_from = fdopen(pipe_b[0], "r");
if(1) fprintf(stderr, "GNU Go started -- %s.\n", __func__);
#endif

  setbuf(stdout, NULL);
  gtp_infile = stdin;
  gtp_outfile = stdout;
  gtp_dumpfile = NULL;
  /* GTP loop */
  gtp_mainloop();

  t = time(NULL);
  i = gg_finish_playout();
  gprintf("Total number of playout=%d\n" "CPU time W=%.1fs B=%.1fs\n"
		  "Elapsed time W=%.1fs B=%.1fs\n",
		  i,
		  gg_colored_clock[GG_C2X(GG_WHITE)].cputime,
		  gg_colored_clock[GG_C2X(GG_BLACK)].cputime,
		  gg_colored_clock[GG_C2X(GG_WHITE)].elapsed,
		  gg_colored_clock[GG_C2X(GG_BLACK)].elapsed);
  fprintf(stderr, "Game end at %s", asctime(localtime(&t)));
#if USE_HASH
  gg_report_hash(stderr);
#endif
exit0:
  rc = 0;
exit:
#if USE_ORACLE
  if(metamachine != 0) {
    kill(metamachine, SIGKILL);
  }
#endif
  return rc;
}
