/***************************************************************
 * mc.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "config.h"
#include "mc.h"

#define MC_TRACE	0

#ifdef __SPU__
int		mc_board_size;
#endif
/****************************************
 *    Macros to include printutils.c	*
 ****************************************/
#define mprintf		mc_mprintf
#define gprintf		mc_gprintf
#define gfprintf	mc_gfprintf
#define vgprintf	mc_vgprintf
#define x2c		mc_x2c

#define is_not_ob	mc_is_not_ob
#define is_onboard	mc_is_onboard
#define is_onboard2	mc_is_onboard2

#define BLACK		MC_BLACK
#define WHITE		MC_WHITE
#define EMPTY		MC_EMPTY
#define EXTER		MC_EXTER
#define PASS_MOVE	MC_PASS_MOVE
#define RESIGN_MOVE	MC_RESIGN_MOVE
#define POS		MC_POS
#define X		MC_X
#define Y		MC_Y

int		mc_gprintf(const char *, ...);
int		mc_mprintf(const char *, ...);
int		mc_gfprintf(FILE *stream, const char *, ...);
void		mc_vgprintf(FILE *stream, const char *format, va_list ap);

#include "printutils.c"
#include "pack.c"

#define MC_FOR_ADJACENT_POINT(q, p, x, d) \
  for((x) = 0; (q) = (p) + mc_direction[(x)], (x) < 8; (x) += (d))

#define MC_FOR_4_ADJACENT_POINT(q, p) \
  for(int _z = 0; (q) = (p) + mc_direction[_z], _z < 8; _z += 2)

#define MC_FOR_8_ADJACENT_POINT(q, p) \
  for(int _z = 0; (q) = (p) + mc_direction[_z], _z < 8; _z += 1)

#define MC_FOR_ALL_POINT(p) \
  for((p) = MC_BOARDMIN; (p) < MC_BOARDMAX; p += 1)

#define MC_FOR_ALL_COLOR(c) \
  for((c) = MC_MIN_COLOR; (c) <= MC_MAX_COLOR;	\
      (c) += MC_MAX_COLOR - MC_MIN_COLOR)

#define MC_FOR_ALL_REN(s) \
for((s) = mc_position->ren[0].next; (s) != 0; (s) = mc_position->ren[(s)].next)

#define MC_FOR_ALL_POINT_IN_REN(p, s)	\
  for((p) = mc_position->ren[(s)].member; \
    (p) != 0; (p) = mc_position->ren_board[(p)].next)

#if USE_EMPTY_LIST
#define MC_FOR_ALL_EMPTY_POINT(p)	\
  for((p) = mc_position->empty_list[0]; \
      (p) != 0; (p) = mc_position->empty_list[p])
#else
#define MC_FOR_ALL_EMPTY_POINT(p)	\
  for((p) = MC_BOARDMIN; (p) < MC_BOARDMAX; p += 1)
#endif

#if USE_ADJACENT_REN
#define MC_BEGIN_FOR_ALL_ADJACENT_REN(t, s)	\
  for((t) = 0; (t) < mc_position->max_ren_free; (t) += 1) {	\
    if(mc_position->adjacent_ren[s][t] == 0)	continue;
#define MC_END_FOR_ALL_ADJACENT_REN	}
#endif

static const int8_t	mc_direction[8] = {
  MC_N_Y, MC_N_Y + MC_N_X, MC_N_X, MC_N_X - MC_N_Y,
  - MC_N_Y, - MC_N_Y - MC_N_X, - MC_N_X, - MC_N_X + MC_N_Y };

static TLS uint16_t	mc_captured_stones;
static TLS mcMove	mc_captured_pos;
static TLS mcRenID	*mc_dirty_ren0, mc_dirty_ren[MC_MAX_REN + 1];
#if USE_LADDER_MODE
static TLS mcRenID	mc_ladder_attack_ren = 0;
#endif

/* For profiling purposes */
#define NOINLINE __attribute__ ((noinline))
#if 0
/* mc */
static void mc_play_and_analyze(mcMove move, mcColor color) NOINLINE;
static mcRenID mc_update_ren(mcMove p, mcColor c) NOINLINE;
static mcMove mc_select_move(mcColor c, mcMove lastmove) NOINLINE;
static int mc_process_proximity(int n, mcColor c, mcMove lastmove) NOINLINE;
static void mc_initialize_allow(void) NOINLINE;
static int mc_process_ladder(int n, mcColor c) NOINLINE;
static bool mc_is_interest(mcMove m, mcColor color) NOINLINE;
static bool mc_is_eye_filling(mcMove m, mcColor c) NOINLINE;
static int mc_final_score(void) NOINLINE;
static inline double mc_random(void) NOINLINE;
/* update_ren */
static void mc_remove_from_empty_list(const mcMove m) NOINLINE;
static INLINE mcMove mc_last_stone(const mcRenID s) NOINLINE;
static INLINE mcRenID mc_create_ren(const mcMove p, const mcColor c) NOINLINE;
static INLINE int mc_collect_ren(const mcRenID s) NOINLINE;
static INLINE int mc_remove_ren(const mcRenID s) NOINLINE;
static INLINE mcRenID mc_join_ren(mcRenID s1, mcRenID s2) NOINLINE;
static INLINE mcRenID mc_add_stone(const mcMove p, const mcRenID s) NOINLINE;
static INLINE int mc_find_libs_ren(const mcRenID s) NOINLINE;
#endif

/*	Marking functions for board, ren_board
 */
static inline void	mc_mark(mcMove p);
static inline bool	mc_is_marked(mcMove p);
static inline void	mc_reset_mark(void);
static inline void	mc_mark_ren(mcRenID s);
static inline bool	mc_is_marked_ren(mcRenID s);
static inline void	mc_reset_renmark(void);

static inline void
mc_mark(mcMove p) { mc_position->mark[p] = mc_position->mark_num; }
static inline bool
mc_is_marked(mcMove p) { return mc_position->mark[p] == mc_position->mark_num; }
static void
mc_do_reset_mark(void) {
  int p;
#if 0
  fprintf(stderr, "mc_do_reset_mark() in %s num=%d\n", __func__,
      mc_position->mark_num);
#endif
  MC_FOR_ALL_POINT(p)	mc_position->mark[p] = 0;
  mc_position->mark_num = 1;
}
static inline void
mc_reset_mark(void) { if(++mc_position->mark_num == 0) mc_do_reset_mark(); }
static inline void

mc_mark_ren(mcRenID s) { mc_position->ren[s].mark = mc_position->mark_ren_num; }
static inline bool
mc_is_marked_ren(mcRenID s)
  { return mc_position->ren[s].mark == mc_position->mark_ren_num; }
static void
mc_do_reset_renmark(void) {
#if 0
  fprintf(stderr, "mc_do_reset_renmark() in %s num=%d\n", __func__,
      mc_position->mark_ren_num);
#endif
  for(int s = 1; s < MC_MAX_REN; s += 1) mc_position->ren[s].mark = 0;
  mc_position->mark_ren_num = 1;
}
static inline void
mc_reset_renmark(void) {
  if(++mc_position->mark_ren_num == 0)	mc_do_reset_renmark();
}

#define MC_CAN_CAPTURE(m, c) \
  (mc_is_opponent(m, c) && mc_libs_ren(mc_id_ren(m)) == 1)

#define MC_NOT_DAMEZUMARI(m, c) \
  (mc_is_friend(m, c) && mc_libs_ren(mc_id_ren(m)) != 1)

#if 1
static inline bool	mc_is_suicide(mcMove m, mcColor c);
static inline bool 
mc_is_suicide(mcMove m, mcColor c) {
  return !mc_is_empty(MC_NORTH(m)) && !mc_is_empty(MC_EAST(m))
    && !mc_is_empty(MC_SOUTH(m)) && !mc_is_empty(MC_WEST(m))
    && !MC_CAN_CAPTURE(MC_NORTH(m), c) && !MC_CAN_CAPTURE(MC_EAST(m), c)
    && !MC_CAN_CAPTURE(MC_SOUTH(m), c) && !MC_CAN_CAPTURE(MC_WEST(m), c)
    && !MC_NOT_DAMEZUMARI(MC_NORTH(m), c) && !MC_NOT_DAMEZUMARI(MC_EAST(m), c)
    && !MC_NOT_DAMEZUMARI(MC_SOUTH(m), c) && !MC_NOT_DAMEZUMARI(MC_WEST(m), c);
}
#else
static bool __attribute__ ((noinline))
mc_is_suicide(mcMove m, mcColor c) {
  return !mc_is_empty(MC_NORTH(m)) && !mc_is_empty(MC_EAST(m))
    && !mc_is_empty(MC_SOUTH(m)) && !mc_is_empty(MC_WEST(m))
    && !MC_CAN_CAPTURE(MC_NORTH(m), c) && !MC_CAN_CAPTURE(MC_EAST(m), c)
    && !MC_CAN_CAPTURE(MC_SOUTH(m), c) && !MC_CAN_CAPTURE(MC_WEST(m), c)
    && !MC_NOT_DAMEZUMARI(MC_NORTH(m), c) && !MC_NOT_DAMEZUMARI(MC_EAST(m), c)
    && !MC_NOT_DAMEZUMARI(MC_SOUTH(m), c) && !MC_NOT_DAMEZUMARI(MC_WEST(m), c);
}
#endif

static inline bool	mc_is_illegal_ko(mcMove m, mcColor c);
static inline bool
mc_is_illegal_ko(mcMove m, mcColor c)
  { return c == mc_position->ko_info.color && m == mc_position->ko_info.point; }

static inline bool	mc_is_legal_empty(mcMove m, mcColor c);
static inline bool
mc_is_legal_empty(mcMove m, mcColor c) {
  assert(mc_is_not_ob(m));
  return !mc_is_illegal_ko(m, c) && !mc_is_suicide(m, c);
}

static inline bool	mc_is_legal(mcMove m, mcColor c);
static inline bool
mc_is_legal(mcMove m, mcColor c) {
  assert(mc_is_not_ob(m));
  return mc_is_empty(m) && !mc_is_illegal_ko(m, c) && !mc_is_suicide(m, c);
}

static inline bool	mc_is_legalx(mcMove m, mcColor c);
static inline bool
mc_is_legalx(mcMove m, mcColor c) {
  if(m == MC_PASS_MOVE || m == MC_RESIGN_MOVE)	return true;
  assert(mc_is_not_ob(m));
  return mc_is_empty(m) && !mc_is_illegal_ko(m, c) && !mc_is_suicide(m, c);
}

static mcRenID	mc_update_ren(mcMove p, mcColor c);

/****************************************
 *    Macros for including SFMT.c	*
 ****************************************/
#define genrand_real2	mc_random
#define gen_rand32	mc_rand32

#define get_idstring	mc_get_idstring
#define get_min_array_size32	mc_get_min_array_size32
#define get_min_array_size64	mc_get_min_array_size64
#define fill_array32	mc_fill_array32
#define fill_array64	mc_fill_array64
#define gen_rand64	mc_gen_rand64
#define init_gen_rand	mc_init_gen_rand
#define init_by_array	mc_init_by_array

#include "SFMT.c"

/****************************************/
static ggScore
mc_final_score(void) {
  int score;
#if FASTEST
  unsigned int p, c;
#else
  mcMove p;
  mcColor c;
#endif

if(0) mc_gprintf("--> %s()\n", __func__);
  /* TODO Use SIMD insn? */
  score = 0;
  MC_FOR_ALL_POINT(p) {
    c = mc_get_color(p);
    if(c == MC_BLACK)		goto black;
    if(c == MC_WHITE)		goto white;
    if(c == MC_EXTER)		continue;
    if(c == MC_EMPTY) {
      c = mc_get_color(MC_NORTH(p));
      if(c == MC_BLACK)		goto black;
      if(c == MC_WHITE)		goto white;
      c = mc_get_color(MC_EAST(p));
      if(c == MC_BLACK)		goto black;
      if(c == MC_WHITE)		goto white;
      c = mc_get_color(MC_SOUTH(p));
      if(c == MC_BLACK)		goto black;
      if(c == MC_WHITE)		goto white;
      c = mc_get_color(MC_WEST(p));
      if(c == MC_BLACK)		goto black;
      if(c == MC_WHITE)		goto white;
      continue;
    }
    MC_FATAL("Impossible color=%d at %1m\n", c, p);
    continue;
black:
    score += 1 + 1;
white:
    score -= 1;
  }
if(0) mc_gprintf("<-- %s score=%+3d\n", __func__, score);
  return score;
}

static void
mc_play_and_analyze(mcMove move, mcColor color) {
  mcRenID self;

if(0) mc_gprintf("--> %s(%1m,%C)\n", __func__, move, color);

  assert(move != MC_PASS_MOVE);
  /* Update board and rens with gathering some info.
   */
  mc_captured_pos = MC_NO_MOVE;
  self = mc_update_ren(move, color);
  /* Update captured stones info. */
  mc_position->bw_captured[MC_C2X(color)] += mc_captured_stones;
  /* Check ko move.
   */
  if(mc_captured_stones == 1 && mc_libs_ren(self) == 1 &&
	    mc_size_ren(self) == 1) {
    assert(mc_captured_pos != MC_NO_MOVE);
    mc_position->ko_info.color = mc_opposit_color(color);
    mc_position->ko_info.point = mc_captured_pos;
if(0) mc_gprintf("--- %s detect ko %1m %C\n", __func__,
	mc_position->ko_info.point, mc_position->ko_info.color);
  } else {
    mc_position->ko_info.color = MC_EMPTY;
  }
if(0) mc_gprintf("<-- %s\n", __func__);
  return;
}

static inline int	mc_count_libs(mcMove p);
static inline int
mc_count_libs(mcMove p) {
  return (mc_is_empty(MC_NORTH(p)) ? 1 : 0)
       + (mc_is_empty(MC_EAST(p)) ? 1 : 0)
       + (mc_is_empty(MC_SOUTH(p)) ? 1 : 0)
       + (mc_is_empty(MC_WEST(p)) ? 1 : 0);
}

static bool
mc_is_eye_filling(mcMove m, mcColor c) {

#if 1

#define MC_CHECK_ORTHOGONAL(c, p) do {	\
    const mcColor _c = mc_get_color(p);	\
    if(_c != (c) && _c != MC_EXTER)	return false;	\
  } while(0)

#define MC_COUNT_DIAGONAL(c, p, ext, opp) do {	\
    const mcColor _c = mc_get_color(p);		\
    (opp) += (_c == c) ? 1 : 0;			\
    (ext) += (_c == MC_EXTER) ? 1 : 0;		\
  } while(0)

  int n_exter, n_opponent;

  if(c == MC_BLACK) {
    MC_CHECK_ORTHOGONAL(MC_BLACK, MC_NORTH(m));
    MC_CHECK_ORTHOGONAL(MC_BLACK, MC_EAST(m));
    MC_CHECK_ORTHOGONAL(MC_BLACK, MC_SOUTH(m));
    MC_CHECK_ORTHOGONAL(MC_BLACK, MC_WEST(m));

    n_exter = 0;
    n_opponent = 0;
    MC_COUNT_DIAGONAL(MC_WHITE, MC_NE(m), n_exter, n_opponent);
    MC_COUNT_DIAGONAL(MC_WHITE, MC_SE(m), n_exter, n_opponent);
    MC_COUNT_DIAGONAL(MC_WHITE, MC_SW(m), n_exter, n_opponent);
    MC_COUNT_DIAGONAL(MC_WHITE, MC_NW(m), n_exter, n_opponent);
    if(n_opponent + (n_exter > 0 ? 1 : 0) >= 2)	return false;
  } else {
    assert(c == MC_WHITE);
    MC_CHECK_ORTHOGONAL(MC_WHITE, MC_NORTH(m));
    MC_CHECK_ORTHOGONAL(MC_WHITE, MC_EAST(m));
    MC_CHECK_ORTHOGONAL(MC_WHITE, MC_SOUTH(m));
    MC_CHECK_ORTHOGONAL(MC_WHITE, MC_WEST(m));

    n_exter = 0;
    n_opponent = 0;
    MC_COUNT_DIAGONAL(MC_BLACK, MC_NE(m), n_exter, n_opponent);
    MC_COUNT_DIAGONAL(MC_BLACK, MC_SE(m), n_exter, n_opponent);
    MC_COUNT_DIAGONAL(MC_BLACK, MC_SW(m), n_exter, n_opponent);
    MC_COUNT_DIAGONAL(MC_BLACK, MC_NW(m), n_exter, n_opponent);
    if(n_opponent + (n_exter > 0 ? 1 : 0) >= 2)	return false;
  }

#else

  union {
    union {
      uint8_t	b[4];
      uint32_t	i;
    } neighbor, diagonal;
    uint64_t i64;
  } freq;

#if 0
#if 0
  for(int i = 0; i < 4; i += 1) {
    freq.neighbor.b[i] = 0;
    freq.diagonal.b[i] = 0;
  }
#elif 0
  freq.neighbor.i = freq.diagonal.i = 0;
#else
  freq.i64 = 0;
#endif
#endif

  freq.neighbor.i = 0;
  freq.neighbor.b[mc_get_color(MC_NORTH(m))]	+= 1;
  if(freq.neighbor.b[c] + freq.neighbor.b[MC_EXTER] <= 0)	return false;
  freq.neighbor.b[mc_get_color(MC_EAST(m))]	+= 1;
  if(freq.neighbor.b[c] + freq.neighbor.b[MC_EXTER] <= 1)	return false;
  freq.neighbor.b[mc_get_color(MC_SOUTH(m))]	+= 1;
  if(freq.neighbor.b[c] + freq.neighbor.b[MC_EXTER] <= 2)	return false;
  freq.neighbor.b[mc_get_color(MC_WEST(m))]	+= 1;
  if(freq.neighbor.b[c] + freq.neighbor.b[MC_EXTER] <= 3)	return false;

  freq.diagonal.i = 0;
  freq.diagonal.b[mc_get_color(MC_NE(m))]	+= 1;
  freq.diagonal.b[mc_get_color(MC_SE(m))]	+= 1;
  freq.diagonal.b[mc_get_color(MC_SW(m))]	+= 1;
  freq.diagonal.b[mc_get_color(MC_NW(m))]	+= 1;
  if((freq.diagonal.b[mc_opposit_color(c)] +
	    (freq.diagonal.b[MC_EXTER] > 0 ? 1 : 0)) >= 2)	return false;

#endif
  return true;
}

#if MOGO

static bool
mc_is_self_atari(mcMove m, mcColor color) {
  int libs;
  bool rc;
  union {
    uint8_t	b[4];
    uint32_t	i;
  } adjacent;
  mcMove q;
  mcRenID s;

if(0) mc_gprintf("<-> %s(%1m, %C)", __func__, m, color);
  assert(mc_is_not_ob(m));
  assert(mc_is_empty(m));
  rc = false;
#if 0
  for(int i = 0; i < 4; i += 1) adjacent.b[i] = 0;
#else
  adjacent.i = 0;
#endif
  MC_FOR_4_ADJACENT_POINT(q, m) {
    adjacent.b[mc_get_color(q)]	+= 1;
  }
  libs = adjacent.b[MC_EMPTY];
  if(libs >= 2)		goto fail;	/* Shortcut */

  /* Can capture? */
  if(adjacent.b[mc_opposit_color(color)] > 0) {
    MC_FOR_4_ADJACENT_POINT(q, m) {
      if(!mc_is_opponent(q, color))	continue;
      s = mc_id_ren(q);
      assert(s != 0);
      if(mc_libs_ren(s) <= 1) {
#if 1
	assert(mc_libs_ren(s) != 0);
#else
	if(mc_libs_ren(s) == 0) {
	  MC_FATAL("libs_ren(%d) at %1m == 0\n", s, q);
	}
#endif
	assert(mc_liberty_ren(s)[0] == m);
      	goto fail;
      }
    }
  }

  /* No. Can connect friends?
   */
  if(adjacent.b[color] > 0) {
#if 1
    mc_reset_renmark();
    mc_reset_mark();
    mc_mark(m);			/* M will not be a liberty. */
    MC_FOR_4_ADJACENT_POINT(q, m) {
      if(!mc_is_friend(q, color))		continue;
      s = mc_id_ren(q);
      assert(s != 0);
      if(mc_is_marked_ren(s))	continue;
      for(int i = 0; i < mc_libs_ren(s); i += 1) {
      	if(i < MC_MAX_LIBS_REN) {
      	  mcMove x = mc_liberty_ren(s)[i];
      	  assert(mc_is_empty(x));
      	  if(mc_is_marked(x))	goto next_liberty;
      	  mc_mark(x);
	}
	libs += 1;
	if(libs >= 2)		goto fail;
next_liberty: ;
      }
      mc_mark_ren(s);
    }
#elif 0				/* This is the old, buggy version */
    mc_reset_renmark();
    MC_FOR_4_ADJACENT_POINT(q, m) {
      if(!mc_is_friend(q, color))		continue;
      s = mc_id_ren(q);
      assert(s != 0);
      if(mc_is_marked_ren(s))	continue;
      libs += mc_libs_ren(s) - 1;
      mc_mark_ren(s);
    }
#else
    /* TODO Exact counting of shared liberties is needed? */
    /* We can use mark_ren(). Which is faster? */
    int j, k;
    mcRenID ren[4];
    k = 0;
    MC_FOR_4_ADJACENT_POINT(q, m) {
      if(!mc_is_friend(q, color))		continue;
      s = mc_id_ren(q);
      assert(s != 0);
      for(j = 0; j < k; j += 1) {
	if(s == ren[j])		goto next_ren;
      }
      libs += mc_libs_ren(s) - 1;
      if(libs >= 2)	goto fail;
      ren[k++] = s;
next_ren:
      continue;
    }
#endif
  } else {
    assert(libs == 1);
  }
  rc = true;
fail:
if(0) mc_gprintf(" %b\n", rc);
  return rc;
}
#endif

/********************************
 *	Select next move	*
 ********************************/

#if ADD_33 != 0
#if MC_UNIQUE
static TLS mcMove	mc_moves[(MC_BOARDSIZE + 3) / 4 * 4 * ADD_33];
#else
static TLS mcMove	mc_moves[(MC_BOARDSIZE + 3) / 4 * 4 * 5 * ADD_33];
#endif
#else
#if MC_UNIQUE
static TLS mcMove	mc_moves[(MC_BOARDSIZE + 3) / 4 * 4];
#else
static TLS mcMove	mc_moves[(MC_BOARDSIZE + 3) / 4 * 4 * 5];
#endif
#endif

#if PURE_RANDOM

static mcMove
mc_select_move(mcColor c, mcMove lastmove) {
  int n;
  mcMove m;

if(0) mc_gprintf("--> %s(%C, %1m)\n", __func__, c, lastmove);
  /* Check legal moves and eye_filling moves.
   */
  n = 0;
  MC_FOR_ALL_EMPTY_POINT(m) {
    if(!mc_is_legal_empty(m, c) || mc_is_eye_filling(m, c))	continue;
    mc_moves[n++] = m;
  }
  /* Select a random move uniformly.
   */
  m = (n == 0 ? MC_PASS_MOVE : mc_moves[(n == 1) ? 0 :
	(int)((double)n * mc_random())]); /* Never use float. */
if(0) mc_gprintf("<-- %s n=%d move=%1m\n", __func__, n, m);
  return m;
}

#else

static const int mc_trans_matrix[8][4] = {
  {  1,  0,  0,  1 },
  {  1,  0,  0, -1 },
  {  0,  1,  1,  0 },
  {  0,  1, -1,  0 },
  {  0, -1,  1,  0 },
  {  0, -1, -1,  0 },
  { -1,  0,  0,  1 },
  { -1,  0,  0, -1 }
};

#if ! USE_PATTERN_TABLE
static INLINE int
mc_transform(int n, int x, int y) {
  int u, v;
  assert((unsigned int)n < 8);
  u = mc_trans_matrix[n][0] * x + mc_trans_matrix[n][1] * y;
  v = mc_trans_matrix[n][2] * x + mc_trans_matrix[n][3] * y;
  return MC_N * v + u;
}
#endif

#if USE_PATTERN_TABLE

#include "pattern.c"

static inline uint16_t
mc_pack_3x3(ggMove p) {
  uint16_t x;
  x = mc_get_color(MC_NE(p));
  x <<= 2; x |= mc_get_color(MC_NORTH(p));
  x <<= 2; x |= mc_get_color(MC_NW(p));
  x <<= 2; x |= mc_get_color(MC_EAST(p));
  x <<= 2; x |= mc_get_color(MC_WEST(p));
  x <<= 2; x |= mc_get_color(MC_SE(p));
  x <<= 2; x |= mc_get_color(MC_SOUTH(p));
  x <<= 2; x |= mc_get_color(MC_SW(p));
  return x;
}

static bool
mc_is_interest(mcMove m, mcColor c) {
  uint16_t x = mc_pack_3x3(m);
  return ((mc_interest_pattern[MC_C2X(c)][x >> 4] & (1 << (x & 0xf))) != 0);
}

#else	/* USE_PATTERN_TABLE */

static inline bool
mc_is_opp_or_ext(mcMove m, mcColor c) {
   return mc_is_opponent(m, c) || mc_is_exter(m);
}

static bool
mc_is_interest(mcMove m, mcColor color) {
  bool rc;
  int i;
  int dx, dy;
  mcColor c;

  rc = true;
  for(i = 0; i < 8; i += 1) {
    dx = mc_transform(i, 1, 0);
    dy = mc_transform(i, 0, 1);
if(0) mc_gprintf("--- %s i=%d dx=%d dy=%d\n", __func__, i, dx, dy);
    if(mc_is_opponent(m + dy, color) &&		/* Hane-4 */
       mc_is_friend(m + dy - dx, color) &&
       mc_is_opponent(m + dy + dx, color) &&
       mc_is_empty(m - dx) && mc_is_empty(m + dx) &&
       mc_is_empty(m - dy))	goto success;
    MC_FOR_ALL_COLOR(c) {
      if(mc_is_opponent(m + dy, c) &&		/* Hane-1 */
	 mc_is_friend(m + dy - dx, c) && mc_is_friend(m + dy + dx, c) &&
	 mc_is_empty(m - dx) && mc_is_empty(m + dx))	goto success;
      if(mc_is_opponent(m + dy, c) &&		/* Hane-2 */
	 mc_is_friend(m + dy - dx, c) && mc_is_empty(m + dy + dx) &&
	 mc_is_empty(m - dx) && mc_is_empty(m + dx) &&
	 mc_is_empty(m - dy))	goto success;
      if(mc_is_opponent(m + dy, c) &&		/* Hane-3 */
	 mc_is_friend(m + dy - dx, c) &&
	 mc_is_friend(m - dx, c) && mc_is_empty(m + dx) &&
	 mc_is_empty(m - dy))	goto success;
      if(mc_is_opponent(m + dy, c) &&		/* Cut-1-1 */
	 mc_is_friend(m + dy - dx, c) &&
	 mc_is_opponent(m - dx, c) &&
	 ! (mc_is_opponent(m + dx, c) &&	/* Cut-1-2 */
	    mc_is_empty(m - dy)) &&
	 ! (mc_is_empty(m + dx) &&		/* Cut-1-3 */
	    mc_is_opponent(m - dy, c)))	goto success;
      if(mc_is_friend(m + dy, c) &&		/* Cut-2 */
	 mc_is_opponent(m - dx, c) && mc_is_opponent(m + dx, c) &&
	 !mc_is_opponent(m - dy, c) &&
	 !mc_is_opponent(m - dy - dx, c) &&
	 !mc_is_opponent(m - dy + dx, c))	goto success;
#if USE_NAKADE_PATTERN && ! USE_PATTERN_TABLE
      if(mc_is_opponent(m + dy - dx, c) &&
	 mc_is_opponent(m + dy, c) &&
	 mc_is_opponent(m + dy + dx, c) &&
      	 mc_is_empty(m - dx) && mc_is_empty(m + dx) &&
	 mc_is_opp_or_ext(m - dy - dx, c) &&
	 mc_is_opp_or_ext(m - dy, c) &&
	 mc_is_opp_or_ext(m - dy + dx, c) &&
	 !mc_is_exter(m - dx) &&
	 mc_is_opp_or_ext(m - dx - dx, c) &&
	 !mc_is_exter(m + dx) &&
	 mc_is_opponent(m + dx + dx, c))	goto success;
      if(mc_is_empty(m + dy) && mc_is_empty(m + dx) &&
	 mc_is_opp_or_ext(m - dy, c) &&
	 mc_is_opp_or_ext(m - dy + dx, c) &&
	 mc_is_opp_or_ext(m - dx, c) &&
	 mc_is_opp_or_ext(m + dy - dx, c) &&
	 !mc_is_exter(m + dy) &&
	 mc_is_opponent(m + dy + dy, c) &&
	 !mc_is_exter(m + dx) &&
	 mc_is_opponent(m + dx + dx, c))	goto success;
#endif
    }
    /* Edge */
    if(mc_is_exter(m - dy) &&
       mc_is_exter(m - dy - dx) && mc_is_exter(m - dy + dx)) {
      MC_FOR_ALL_COLOR(c) {
        if(mc_is_empty(m + dy) &&		/* Edge-1 */
           mc_is_friend(m + dy - dx, c) &&
           mc_is_opponent(m - dx, c))	goto success;
        if(mc_is_friend(m + dy, c) &&		/* Edge-2 */
           !mc_is_friend(m - dx, c) && mc_is_opponent(m + dx, c)) goto success;
      }
      if(mc_is_friend(m + dy, color) &&		/* Edge-3 */
	 mc_is_opponent(m + dy + dx, color))	goto success;
      if(mc_is_opponent(m + dy, color) &&	/* Edge-4 */
      	 mc_is_friend(m + dy + dx, color) &&
	 !mc_is_opponent(m + dx, color))	goto success;
      if(mc_is_opponent(m + dy, color) &&	/* Edge-5 */
      	 mc_is_friend(m + dy + dx, color) &&
	 mc_is_friend(m - dx, color) && mc_is_opponent(m + dx, color))
	goto success;
    }
  }
  rc = false;
success:
  return rc;
}

#endif	/* USE_PATTERN_TABLE */

#if GENERATE_PATTERN

static uint16_t
mc_pack16(uint8_t *p) {
  uint16_t x = 0;
  for(int i = 0; i < 0x10; i += 1) {
    x <<= 1; x |= (*p++ != 0);
  }
  return x;
}

void	mc_print_pattern(void);
void
mc_print_pattern(void) {
  unsigned int i, p;
  ggColor c;
  uint8_t pattern[2][0x10000];

  /* Generate all possible 3x3 patterns for both colors.
   */
  mc_set_color(MC_POS(1, 1), MC_EMPTY);
  MC_FOR_ALL_COLOR(c) {
    for(i = 0; i < 0x10000; i += 1) {
      p = i;
      mc_set_color(MC_POS(0, 0), p & 0x3); p >>= 2;
      mc_set_color(MC_POS(1, 0), p & 0x3); p >>= 2;
      mc_set_color(MC_POS(2, 0), p & 0x3); p >>= 2;
      mc_set_color(MC_POS(0, 1), p & 0x3); p >>= 2;
      mc_set_color(MC_POS(2, 1), p & 0x3); p >>= 2;
      mc_set_color(MC_POS(0, 2), p & 0x3); p >>= 2;
      mc_set_color(MC_POS(1, 2), p & 0x3); p >>= 2;
      mc_set_color(MC_POS(1, 2), p & 0x3); p >>= 2;
      pattern[MC_C2X(c)][i] = mc_is_interest(MC_POS(1, 1), c);
    }
  }
  printf("static uint16_t mc_interest_pattern[2][0x1000] = {");
  MC_FOR_ALL_COLOR(c) {
    printf(" {");
    i = 0; do {
      if(i % 8 == 0)	printf("\n\t"); 
      printf("%#x", mc_pack16(&pattern[MC_C2X(c)][0x10 * i]));
      if(++i >= 0x1000)	break;
      printf(",");
    } while(1);
    printf(" },\n");
  }
  printf("};\n");
  return;
}

#endif	/* GENERATE_PATTERN */

/* Check legal moves and eye_filling moves.
 */
static TLS uint8_t	mc_allow_cache[(MC_BOARDSIZE + 3) / 4 * 4];

static inline bool
mc_allow(mcMove m, mcColor c) {
  if(mc_allow_cache[m] >= 2) {
    return true;
  } else if(mc_allow_cache[m] == 0) {
    if(mc_is_legal(m, c) && !mc_is_eye_filling(m, c)) {
      mc_allow_cache[m] = 2;
      return true;
    }
    mc_allow_cache[m] = 1;
    return false;
  } else {
    return false;
  }
}

#if MC_UNIQUE
#define PUSH_MOVE(n, m)	do {	\
  if(mc_allow_cache[m] != 3) {	\
    mc_allow_cache[m] = 3;	\
    mc_moves[(n)++] = (m);	\
  }				\
} while(0)
#else
#define PUSH_MOVE(n, m)		\
  mc_moves[(n)++] = (m)
#endif

#define PUSH_MOVE_WITH_CHECK(n, m, c)	\
  if(mc_allow((m), (c))) PUSH_MOVE((n), (m))

static void
mc_initialize_allow(void) {
#if 0
  mcMove m;
  for(m = 0; m < (MC_BOARDSIZE + 3) / 4 * 4; m += 1) { mc_allow_cache[m] = true; }
#else
  memset(&mc_allow_cache[0], 0, sizeof mc_allow_cache);
#endif
  return;
}

#if USE_LADDER_MODE

static TLS mcBool	ladder_attack_pos[MC_BOARDSIZE];
static TLS bool		ladder_attack_possible = false;
static TLS mcRenID	ladder_attack_ren = 0;

/* Process ladder
 */

/* Try NOBI or TSUGI. */
static bool
mc_can_escape(mcMove m, mcColor c, int libs, int threshold) {
  mcMove p;
  mcRenID s;

  if(mc_count_libs(m) - 1 + libs >= threshold)	return true;
  MC_FOR_4_ADJACENT_POINT(p, m) {
    if(!mc_is_friend(p, c))	continue;
    s = mc_id_ren(p);
    assert(s != 0);
    if(mc_libs_ren(s) - 2 + libs >= threshold)	return true;
  }
  return false;
}

#if 1

static int
mc_process_ladder(int n, mcColor c) {
  const int n0 = n;
  int libs;
  mcMove m, p, q;
  mcRenID s, t;

  s = mc_ladder_attack_ren;
  if(s != 0) {
    /* Already captured? */
    if(mc_color_ren(s) == MC_EMPTY)	goto stop_ladder_mode;
    libs = mc_libs_ren(s);
    assert(libs > 0);
    if(libs >= 3)			goto stop_ladder_mode;
    if(mc_color_ren(s) == c) {		/* My ren is being captured. */
      /* Try NOBI or TSUGI.
       */
      for(int i = 0; i < libs; i += 1) {
      	m = mc_liberty_ren(s)[i];
      	if(!mc_allow(m, c))		continue;
      	if(!mc_can_escape(m, c, libs, 2))	continue;
      	if(mc_is_self_atari(m, c))	continue;	/* Necessary? */
	PUSH_MOVE(n, m);
      }
      /* Try capture opponent rens.
       */
      /* MC_FOR_ALL_ADJACENT_REN(t, s) */
      mc_reset_renmark();
      MC_FOR_ALL_POINT_IN_REN(p, s) {
	MC_FOR_4_ADJACENT_POINT(q, p) {
	  if(!mc_is_opponent(q, c))	continue;
	  t = mc_id_ren(q);
	  assert(t != 0);
	  if(mc_is_marked_ren(t))	continue;
	  mc_mark_ren(t);
	  if(mc_libs_ren(t) >= 2)	continue;
	  assert(mc_libs_ren(t) == 1);
	  m = mc_liberty_ren(t)[0];
	  PUSH_MOVE_WITH_CHECK(n, m, c);
	}
      }
    } else {				/* Capturing an opponent ren. */
      assert(mc_color_ren(s) == mc_opposit_color(c));
      if(libs == 1) {
	/* Capture! */
	m = mc_liberty_ren(s)[0];
	PUSH_MOVE_WITH_CHECK(n, m, c);
      } else if(libs == 2) {		/* Continue attack? */
	/* Loose ladder? */
	  /* MC_FOR_ALL_ADJACENT_REN(t, s) */
	MC_FOR_ALL_POINT_IN_REN(p, s) {
	  MC_FOR_4_ADJACENT_POINT(q, p) {
	    if(!mc_is_friend(q, c))	continue;
	    t = mc_id_ren(q);
	    assert(t != 0);
	    if(mc_libs_ren(t) == 1)	goto stop_ladder_mode;
	    assert(mc_libs_ren(t) != 0);
	  }
	}
	/* Continue attack.
	 */
{
	mcMove m0 = mc_liberty_ren(s)[0];
	mcMove m1 = mc_liberty_ren(s)[1];
	int libs0 = mc_count_libs(m0);
	int libs1 = mc_count_libs(m1);

	m = m1;		/* To avoid compiler warning. */
	if(!mc_allow(m0, c) || mc_is_self_atari(m0, c) ||
	    mc_can_escape(m1, mc_opposit_color(c), libs, 2)) {
	  libs0 = -1;
//	  m = m1;	/* Already done. */
//	  m0 = m0;
	}
	if(!mc_allow(m1, c) || mc_is_self_atari(m1, c) ||
	    mc_can_escape(m0, mc_opposit_color(c), libs, 2)) {
	  libs1 = -1;
	  m = m0;
	  m0 = m1;
	}
	if(libs0 < 0 && libs1 < 0)	goto stop_ladder_mode;
	if(libs0 < 0 || libs1 < 0) {
	  libs = mc_count_libs(m0);
	  if(libs > 2)			goto stop_ladder_mode;
	  assert(mc_allow(m, c));
	  PUSH_MOVE(n, m);
	} else {
	  assert(mc_allow(m0, c));
	  assert(mc_allow(m1, c));
	  if(libs0 > 2 && libs1 > 2)	goto stop_ladder_mode;
	  else if(libs0 > 2) {
	    PUSH_MOVE(n, m0);
	  } else if(libs1 > 2) {
	    PUSH_MOVE(n, m1);
	  } else {
	    PUSH_MOVE(n, m0);
	    PUSH_MOVE(n, m1);
	  }
	}
}
      }
    }
    if(n > n0) goto exit;
stop_ladder_mode:
    mc_ladder_attack_ren = 0;
  }

/* Mark-up all ladder attack points.
 */
if(0) mc_gprintf("--> %s (Mark-ladder)\n", __func__);
  ladder_attack_possible = false;
  MC_FOR_ALL_EMPTY_POINT(m) {
    ladder_attack_pos[m] = false;
  }
  /* Seek rens with 2 liberties. */
  MC_FOR_ALL_REN(s) {
    assert(s != 0);
    if(mc_color_ren(s) == c)		continue;
    assert(mc_color_ren(s) == mc_opposit_color(c));
    libs = mc_libs_ren(s);
    assert(libs != 0);
    if(libs != 2)		continue;
    /* Check loose ladder ? */
    /* MC_FOR_ALL_ADJACENT_REN(t, s) */
    mc_reset_renmark();
    MC_FOR_ALL_POINT_IN_REN(p, s) {
      MC_FOR_4_ADJACENT_POINT(q, p) {
	if(!mc_is_friend(q, c))		continue;
	t = mc_id_ren(q);
	assert(t != 0);
	if(mc_is_marked_ren(t))		continue;
	mc_mark_ren(t);
	if(mc_libs_ren(t) <= 1)		goto next_ren;
	for(int i = 0; i < 2; i += 1) {
	  m = mc_liberty_ren(s)[i];
	  if(!mc_allow(m, c))		continue;
	  if(mc_is_self_atari(m, c))	continue;
	  ladder_attack_possible = true;
	  ladder_attack_pos[m] = true;
	}
	goto next_ren;
      }
    }
next_ren:
    continue;
  }
exit:
  return n;
}

#else

static int
mc_process_ladder(int n, mcColor c) {
  const int n0 = n;
  int libs;
  mcMove m, p, q;
  mcRenID s, t;

  /* During ladder attack?
   */
  if((s = mc_ladder_attack_ren) != 0) {
    /* Already captured.  */
    if(mc_color_ren(s) == MC_EMPTY)	goto stop_ladder_mode;
    if(mc_color_ren(s) == c) {
      /* Friends are being captured. */
      if(mc_libs_ren(s) > 1)	goto stop_ladder_mode;
      /* Try escape.
       */
      n = 0;
      mc_reset_mark();
      /* Try NOBI or TSUGI.
       * Push all liberties of the ren except self-atari points.
       */
      MC_FOR_ALL_POINT_IN_REN(p, s) {
	MC_FOR_4_ADJACENT_POINT(m, p) {
	  if(!mc_allow(m, c))		continue;
	  if(mc_is_self_atari(m, c))	continue;
	  PUSH_MOVE(n, m);
	}
      }
      /* Try capture opponent rens.
       */
      /* MC_FOR_ALL_ADJACENT_REN(t, s) is much better. */
      mc_reset_renmark();
      MC_FOR_ALL_POINT_IN_REN(p, s) {
	MC_FOR_4_ADJACENT_POINT(q, p) {
	  if(!mc_is_opponent(q, c))	continue;
	  t = mc_id_ren(q);
	  assert(t != 0);
	  if(mc_is_marked_ren(t))	continue;
	  mc_mark_ren(t);
	  if(mc_libs_ren(t) > 1)	continue;
	  m = mc_liberty_ren(t)[0];
	  if(!mc_allow(m, c))			continue;
	  PUSH_MOVE(n, m);
	}
      }
    } else {
      assert(mc_color_ren(s) == mc_opposit_color(c));
      /* Continue attack? */
      libs = mc_libs_ren(s);
      if(libs == 1) {
	/* Capture! */
	m = mc_liberty_ren(s)[0];
	assert(mc_allow(m, c));
	PUSH_MOVE(n, m);
      } else if(libs == 2) {
	/* Loose ladder? */
	  /* MC_FOR_ALL_ADJACENT_REN(t, s) */
	mc_reset_renmark();
	MC_FOR_ALL_POINT_IN_REN(p, s) {
	  MC_FOR_4_ADJACENT_POINT(q, p) {
	    if(!mc_is_friend(q, c))	continue;
	    t = mc_id_ren(q);
	    assert(t != 0);
	    if(mc_is_marked_ren(t))	continue;
	    mc_mark_ren(t);
	    if(mc_libs_ren(t) < 2)	goto stop_ladder_mode;
	  }
	}
	/* Continue attack.
	 */
	m = mc_liberty_ren(s)[0];
	if(!mc_allow(m, c) || mc_is_self_atari(m, c)) {
	  m = mc_liberty_ren(s)[1];
	  if(!mc_allow(m, c) || mc_is_self_atari(m, c)) goto stop_ladder_mode;
	} else {
	  m = mc_liberty_ren(s)[1];
	  if(!mc_allow(m, c) || mc_is_self_atari(m, c)) {
	    m = mc_liberty_ren(s)[0];
	  } else {
	    m = mc_liberty_ren(s)[(int)(2 * mc_random())];
	  }
	}
      } else {
	/* Give up attack. Clear ladder attack mode.
	 */
	assert(mc_libs_ren(s) > 2);
	goto stop_ladder_mode;
      }
    }
    if(n > n0)	goto exit;
stop_ladder_mode:
    mc_ladder_attack_ren = 0;
  }

/* List-up ladder attack points.
 */
if(0) mc_gprintf("--> %s (Listup-ladder)\n", __func__);
  n = n0;
  MC_FOR_ALL_REN(s) {
    assert(s != 0);
    if(mc_color_ren(s) == c)	continue;
    assert(mc_color_ren(s) == mc_opposit_color(c));
    if(mc_libs_ren(s) != 2)	continue;
    /* Loose ladder ? */
    /* MC_FOR_ALL_ADJACENT_REN(t, s) */
    MC_FOR_ALL_POINT_IN_REN(p, s) {
      MC_FOR_4_ADJACENT_POINT(q, p) {
	if(!mc_is_friend(q, c))	continue;
	t = mc_id_ren(q);
	assert(t != 0);
	if(mc_libs_ren(t) < 2)	goto next_ren;
      }
    }
#if 0
    for(i = 0; i < 2; i += 1) {
      m = mc_liberty_ren(s)[i];
      assert(mc_is_not_ob(m));
      if(!mc_allow(m, c))	goto next_ren;
      if(mc_is_self_atari(m, c))	goto next_ren;
    }
#endif
    ladder_attack_possible = true;
    MC_FOR_ALL_EMPTY_POINT(m) {
      if(!mc_allow(m, c))	continue;
      ladder_attack_pos[m] = false;
    }
    for(int i = 0; i < 2; i += 1) {
      m = mc_liberty_ren(s)[i];
      assert(mc_is_not_ob(m));
      if(!mc_allow(m, c))			continue;
      if(mc_is_self_atari(m, c))	continue;
      ladder_attack_pos[m] = true;
    }
next_ren:
    continue;
  }
exit:
  return n;
}

#endif

#endif	/* USE_LADDER_MODE */

#if PRIOR_UTTEGAESHI || PROHIBIT_UTTEGAESHI
static bool
mc_is_uttegaeshi(const mcMove m, const mcRenID t, const mcColor color) {
  mcMove p;
  mcColor c;
  int n_friends;

  if(mc_size_ren(t) == 1) {
    n_friends = 0;
    MC_FOR_4_ADJACENT_POINT(p, m) {
      c = mc_get_color(p);
      if(c == MC_EMPTY)	break;
      n_friends += (c == color) ? 1 : 0;
    }
    if(n_friends == 1)	return true;
  }
  return false;
}
#endif

#if CAPTURE_IS_INTEREST
static bool
mc_is_capture_move(mcMove m, mcColor c, mcMove last) {
  mcRenID s;

  if(mc_get_color(last) != mc_opposit_color(c))	return false;
  s = mc_id_ren(last);
  assert(s != 0);
  if(mc_libs_ren(s) > 1)	return false;
  assert(mc_libs_ren(s) == 1);
  return mc_liberty_ren(s)[0] == m;
}
#endif

#if 0

static int
mc_process_proximity(int n, mcColor c, mcMove lastmove) {
  const int n0 = n;
  int libs;
  mcMove m, p, q, r;
  mcRenID s, t;

if(0) mc_gprintf("--> %s(%d,%C,%1m)\n", __func__, n, c, lastmove);

  assert(mc_is_onboard(lastmove));
  assert(mc_get_color(lastmove) == mc_opposit_color(c));

  /* First, test if last move is atari.
   * If so, select one of escaping moves by means of
   * increasing liberties or capturing adjacent opponents.
   */
if(0) mc_gprintf("--> %s (MOGO-1: Check if last move is atari)\n", __func__);
  /* Assuming suicide is not allowed. */
  MC_FOR_4_ADJACENT_POINT(p, lastmove) {
    if(mc_get_color(p) != c)		continue;
    s = mc_id_ren(p);
    assert(s != 0);
    libs = mc_libs_ren(s);
    if(libs > 1)		continue;
    assert(libs == 1);

    /* 'S' is the ren in atari.
     */
    /* Try extend or connect. */
    m = mc_liberty_ren(s)[0];
    assert(mc_is_not_ob(m));
    assert(mc_is_empty(m));
    /* Can increase liberties or connect to other friendly rens? */
    if(mc_allow(m, c) && mc_can_escape(m, c, libs, 2)) {
      PUSH_MOVE(n, m);
    }
    /* Can capture opponent ren(s)?
     */
    /* TODO MC_FOR_ALL_ADJACENT_REN(t, s) */
    mc_reset_renmark();
    MC_FOR_ALL_POINT_IN_REN(q, s) {
      MC_FOR_4_ADJACENT_POINT(r, q) {
	if(!mc_is_opponent(r, c))	continue;
	t = mc_id_ren(r);
	assert(t != 0);
	if(mc_is_marked_ren(t))		continue;
	mc_mark_ren(t);
	if(mc_libs_ren(t) > 1)		continue;
	assert(mc_libs_ren(t) == 1);
	m = mc_liberty_ren(t)[0];
	assert(mc_is_not_ob(m));
	assert(mc_is_empty(m));
	if(!mc_allow(m, c))		continue;
#if PROHIBIT_UTTEGAESHI
	if(mc_is_uttegaeshi(m, t, c))	continue;
#endif
#if PRIOR_UTTEGAESHI
	if(mc_is_uttegaeshi(m, t, c)) {
	  PUSH_MOVE(n, m);
	}
#endif
	PUSH_MOVE(n, m);
      }
    }
  }
#if DETERMINISTIC
  if(n > n0
#if  KO_THREAT != 0
      && mc_position->ko_info.color == MC_EMPTY
#endif
  )	goto exit;
#endif

/* Second, select one of interest moves around last move.
 */
if(0) mc_gprintf("--> %s (MOGO-2: Interest moves around last move)\n", __func__);
  MC_FOR_8_ADJACENT_POINT(m, lastmove) {
    if(mc_allow(m, c) &&
      (mc_is_interest(m, c)
#if CAPTURE_IS_INTEREST
    	  || mc_is_capture_move(m, c, lastmove)
#endif
	) &&
    	!mc_is_self_atari(m, c)) {
      PUSH_MOVE(n, m);
if(MC_TRACE) mc_gprintf(" ! %1m", m);
    }
  }

#if DETERMINISTIC
exit:
#endif
  return n;
}

#else

static int
mc_process_proximity(int n, mcColor c, mcMove lastmove) {
  const int n0 = n;
  int libs;
  mcMove m, p, q, r;
  mcRenID s, t;

if(0) mc_gprintf("--> %s(%d,%C,%1m)\n", __func__, n, c, lastmove);

  assert(mc_is_onboard(lastmove));
  assert(mc_get_color(lastmove) == mc_opposit_color(c));

  /* First, test if last move is atari.
   * If so, select one of escaping moves by means of
   * increasing liberties or capturing adjacent opponents.
   */
if(0) mc_gprintf("--> %s (MOGO-1)\n", __func__);
  assert(mc_is_not_ob(lastmove));

  /* Asuming suicide is not allowed. */
  MC_FOR_4_ADJACENT_POINT(p, lastmove) {
    if(mc_get_color(p) != c)	continue;
    s = mc_id_ren(p);
    assert(s != 0);
    if(mc_libs_ren(s) >= 2)	continue;
    /* 's' is ATARI. */
    assert(mc_libs_ren(s) == 1);
    m = mc_liberty_ren(s)[0];
    assert(mc_is_not_ob(m));
    assert(mc_is_empty(m));
    if(!mc_allow(m, c))		continue;
    libs = mc_count_libs(m);
    /* Can escape? */
    if(libs < 3) {
      /* Can connect to another friendly ren with more liberties?
       */
      MC_FOR_4_ADJACENT_POINT(r, m) {
	if(!mc_is_friend(r, c))	continue;
	t = mc_id_ren(r);
	assert(t != 0);
	if(t == s)			continue;
#if 0
	if(mc_libs_ren(t) >= 2)	goto add_if;
#else
	if(mc_libs_ren(t) >= 3)	goto add_if;
#endif
      }
#if 0
      continue;
#else
      goto try_capture;
#endif
    }
add_if:
    PUSH_MOVE(n, m);
    /* Can capture opponent ren(s)?
     */
    /* TODO MC_FOR_ALL_ADJACENT_REN(t, s) is much better. */
#if 1
try_capture:
#endif
    mc_reset_renmark();
    MC_FOR_ALL_POINT_IN_REN(q, s) {
      MC_FOR_4_ADJACENT_POINT(r, q) {
	if(!mc_is_opponent(r, c))	continue;
	t = mc_id_ren(r);
	assert(t != 0);
	if(mc_libs_ren(t) > 1)	continue;
	assert(mc_libs_ren(t) == 1);
	if(mc_is_marked_ren(t))	continue;
	mc_mark_ren(t);
	m = mc_liberty_ren(t)[0];
	if(!mc_allow(m, c))			continue;
	PUSH_MOVE(n, m);
      }
    }
  }
#if DETERMINISTIC
  if(n > n0
#if  KO_THREAT != 0
      && mc_position->ko_info.point == MC_NO_MOVE
#endif
  )	goto exit;
#endif

/* Second, select one of interest moves around last move.
 */
if(0) mc_gprintf("--> %s (MOGO-2)\n", __func__);
  MC_FOR_8_ADJACENT_POINT(m, lastmove) {
    if(!mc_allow(m, c))			continue;
    if(mc_is_self_atari(m, c))		continue;
    if(!mc_is_interest(m, c))		continue;
    PUSH_MOVE(n, m);
if(0) mc_gprintf(" !%1m", m);
  }
#if DETERMINISTIC
exit:
#endif
  return n;
}

#endif

static mcMove
mc_select_move(mcColor c, mcMove lastmove) {
  int n;
  mcMove m;
  mcRenID s;

if(0) mc_gprintf("--> %s(%C, %1m)\n", __func__, c, lastmove);

  n = 0;
  mc_initialize_allow();

#if USE_LADDER_MODE
  /* During ladder attack?
   */
  if(mc_position->use_ladder_mode) {
    n = mc_process_ladder(n, c);
    if(n > 0) {
      m = (n == 1) ? mc_moves[0] : mc_moves[(int)((double)n * mc_random())];
      goto exit_wo_check;
    }
  }
#endif

#if PRIOR_KO_TSUGI
  if(mc_position->ko_info.point != MC_NO_MOVE &&
      mc_position->ko_info.color == MC_EMPTY) {
    PUSH_MOVE_WITH_CHECK(n, mc_position->ko_info.point, c);
  }
#endif

#if KO_THREAT == 1
  if(mc_position->ko_info.color != MC_EMPTY) {
    /* Seek atari moves. */
    MC_FOR_ALL_REN(s) {
      assert(s != 0);
      if(mc_color_ren(s) == c)	continue;
      assert(mc_color_ren(s) == mc_opposit_color(c));
      if(mc_libs_ren(s) != 2)	continue;
      m = mc_liberty_ren(s)[0];
      PUSH_MOVE_WITH_CHECK(n, m, c);
      m = mc_liberty_ren(s)[1];
      PUSH_MOVE_WITH_CHECK(n, m, c);
    }
#if DETERMINISTIC
    if(n > 0)		goto select_one_and_exit;
#endif
  }
#elif KO_THREAT == 2
  goto uniform_random;
#endif

  assert(mc_is_not_ob(lastmove));
  if(mc_is_onboard(lastmove)) {
    n = mc_process_proximity(n, c, lastmove);
#if DETERMINISTIC
    if(n > 0)		goto select_one_and_exit;
#endif
  }

  /* Third, seek capturing moves.
   */
if(0) mc_gprintf("--> %s (MOGO-3: Capturing moves)\n", __func__);

  MC_FOR_ALL_REN(s) {
    assert(s != 0);
    if(mc_color_ren(s) == c)	continue;
    assert(mc_color_ren(s) == mc_opposit_color(c));
    if(mc_libs_ren(s) > 1)	continue;
    assert(mc_libs_ren(s) == 1);
    m = mc_liberty_ren(s)[0];
    assert(mc_is_onboard(m));
#if 1
    PUSH_MOVE_WITH_CHECK(n, m, c);
#else
    PUSH_MOVE(n, m);
#endif
  }
#if DETERMINISTIC
  if(n > 0)	goto select_one_and_exit;
#endif

#if KO_THREAT == 2
uniform_random:
#endif
/* Last, select a move randomly.
 */
if(0) mc_gprintf("--> %s (MOGO-4: Random)\n", __func__);
  n = 0;
  MC_FOR_ALL_EMPTY_POINT(m) {
    if(!mc_allow(m, c))	continue;
    mc_moves[n++] = m;
  }

  if(n == 0) {
    m = MC_PASS_MOVE;
    goto exit_wo_check;
  }

#define MC_ADD_MOVE(n, m, c, k) do {	\
  if(mc_allow(m, c)) {			\
    for(int i = 0; i < (k); i += 1) {	\
      mc_moves[(n)++] = (m);		\
    }					\
  }					\
} while(0)

#if ADD_33 != 0
  MC_ADD_MOVE(n, MC_POS(2, 2), c, ADD_33);
  MC_ADD_MOVE(n, MC_POS(2, mc_board_size - 3), c, ADD_33);
  MC_ADD_MOVE(n, MC_POS(mc_board_size - 3, 2), c, ADD_33);
  MC_ADD_MOVE(n, MC_POS(mc_board_size - 3, mc_board_size - 3), c, ADD_33);
#endif

#if ADD_21 != 0
  MC_ADD_MOVE(n, MC_POS(1, 0), c, ADD_21);
  MC_ADD_MOVE(n, MC_POS(0, 1), c, ADD_21);
  MC_ADD_MOVE(n, MC_POS(mc_board_size - 2, 0), c, ADD_21);
  MC_ADD_MOVE(n, MC_POS(mc_board_size - 1, 1), c, ADD_21);
  MC_ADD_MOVE(n, MC_POS(1, mc_board_size - 1), c, ADD_21);
  MC_ADD_MOVE(n, MC_POS(0, mc_board_size - 2), c, ADD_21);
  MC_ADD_MOVE(n, MC_POS(mc_board_size - 1, mc_board_size - 2), c, ADD_21);
  MC_ADD_MOVE(n, MC_POS(mc_board_size - 2, mc_board_size - 1), c, ADD_21);
#endif

#if ADD_CAPTURE_SIZE != 0
  MC_FOR_ALL_REN(s) {
    assert(s != 0);
    if(mc_color_ren(s) == c)	continue;
    assert(mc_color_ren(s) == mc_opposit_color(c));
    if(mc_libs_ren(s) > 1)	continue;
    assert(mc_libs_ren(s) == 1);
    if(mc_size_ren(s) < ADD_CAPTURE_SIZE)	continue;
    m = mc_liberty_ren(s)[0];
    MC_ADD_MOVE(n, m, c, ADD_CAPTURE);
  }
#endif

#if ADD_NAKADE != 0
  MC_FOR_ALL_EMPTY_POINT(m) {
    for(int i = 0; i < 8; i += 1) {
      mcColor d;
      int dx = mc_transform(i, 1, 0);
      int dy = mc_transform(i, 0, 1);
      MC_FOR_ALL_COLOR(d) {
	if(mc_is_opponent(m + dy - dx, d) &&
	   mc_is_opponent(m + dy, d) &&
	   mc_is_opponent(m + dy + dx, d) &&
	   mc_is_empty(m - dx) && mc_is_empty(m + dx) &&
	   mc_is_opp_or_ext(m - dy - dx, d) &&
	   mc_is_opp_or_ext(m - dy, d) &&
	   mc_is_opp_or_ext(m - dy + dx, d) &&
	   !mc_is_exter(m - dx) &&
	   mc_is_opp_or_ext(m - dx - dx, d) &&
	   !mc_is_exter(m + dx) &&
	   mc_is_opponent(m + dx + dx, d))	goto match;
	if(mc_is_empty(m + dy) && mc_is_empty(m + dx) &&
	   mc_is_opp_or_ext(m - dy, d) &&
	   mc_is_opp_or_ext(m - dy + dx, d) &&
	   mc_is_opp_or_ext(m - dx, d) &&
	   mc_is_opp_or_ext(m + dy - dx, d) &&
	   !mc_is_exter(m + dy) &&
	   mc_is_opponent(m + dy + dy, d) &&
	   !mc_is_exter(m + dx) &&
	   mc_is_opponent(m + dx + dx, d))	goto match;
      }
      continue;
match:
      if(mc_allow(m, c)) {
      	MC_ADD_MOVE(n, m, c, ADD_NAKADE);
      }
      break;
    }
  }
#endif

#if DETERMINISTIC
select_one_and_exit:
#endif
  m = (n == 1) ? mc_moves[0] : mc_moves[(int)((double)n * mc_random())];

#if USE_LADDER_MODE
  if(mc_position->use_ladder_mode) {
    if(mc_ladder_attack_ren == 0 && ladder_attack_possible &&
	ladder_attack_pos[m]) {
      mc_ladder_attack_ren = ladder_attack_ren;
if(0) mc_gprintf("--- %s (LADDER) move=%1m\n", __func__, m);
    }
  }
#endif
exit_wo_check:
  assert(mc_is_legalx(m, c));
if(0) mc_gprintf("<-- %s move=%1m\n", __func__,  m);
  return m;
}

#endif	/* MOGO */

TLS mcPosition	*mc_position;

ggScore
mc(mcPosition *p, mcRave *r, mcMove move) {
  ggScore score;
  mcColor color;
#if USE_RAVE
  int n_rave;
#endif
#if ADAPT_JAPANESE
  mcColor last_played_color = MC_EMPTY;
#endif

if(0) mc_gprintf("\n--> %s(%p,%p,%1m)", __func__, p, r, move);
  mc_position = p;
#if USE_LADDER_MODE
  mc_ladder_attack_ren = 0;
#endif
#if USE_RAVE
  n_rave = 0;
#endif
  assert(p->n_pass < 2);
  assert(!p->gameend);
  color = p->color;
  do {
    mc_captured_pos = MC_NO_MOVE;
if(MC_TRACE) mc_gprintf(" %1m(%C)", move, color);
    if(move == MC_PASS_MOVE) {
      if(++p->n_pass >= 2)		goto gameend;
      p->ko_info.color = MC_EMPTY;
    } else {
#if ADAPT_JAPANESE
      last_played_color = color;
#endif
      p->n_pass = 0;
      mc_play_and_analyze(move, color);
      if(mc_captured_pos != MC_NO_MOVE) {
#if USE_CAPTURED == 1
	score = p->bw_captured[MC_C2X(MC_BLACK)]
	      - p->bw_captured[MC_C2X(MC_WHITE)];
	if(score > CAPTURED_THRESHOLD) {
	  score = CAPTURED_THRESHOLD;
	  goto mercy;
	} else if(- score > CAPTURED_THRESHOLD) {
	  score = - CAPTURED_THRESHOLD;
	  goto mercy;
	}
#elif USE_CAPTURED == 2
{
	int cap1 = p->bw_captured[MC_C2X(MC_BLACK)],
	    cap2 = p->bw_captured[MC_C2X(MC_WHITE)];
	if(cap1 > 2 * cap2 && cap2 > CAPTURED_THRESHOLD) {
	  score = CAPTURED_THRESHOLD;
	  goto mercy;
	} else if(cap2 > 2 * cap1 && cap1 > CAPTURED_THRESHOLD) {
	  score = - CAPTURED_THRESHOLD;
	  goto mercy;
	}
}
#endif
      }
    }
    if(++p->movenum >= MC_MAX_MOVENUM)	break;
    color = mc_opposit_color(color);
    move = mc_select_move(color, move);
#if USE_RAVE
    assert(n_rave < MC_MAX_MOVENUM);
    r->move[n_rave] = move;
    r->color[n_rave] = color;
    n_rave += 1;
#endif
  } while(true);
if(0) mc_gprintf("Moves exceeded MAX_MOVENUM (%d) -- %s\n",
	      p->movenum, __func__);
  score = GG_UNKNOWN_SCORE;
  goto exit;

gameend:
  p->gameend = true;
  score = mc_final_score();
#if ADAPT_JAPANESE
  score -= (last_played_color == MC_BLACK) ? 1 : 0;
#endif
mercy:
  if(p->color != MC_BLACK) {
    score = - score;
  }
  goto exit;

exit:
if(MC_TRACE) mc_gprintf("\n");
#if USE_RAVE
  r->n_rave = n_rave;
#endif
if(0) mc_gprintf("<-- %s() score=%d\n", __func__, score);
  return score;
}

/****************************************
 *    Macros to include update_ren.c	*
 ****************************************/
#define Color		mcColor
#define Move		mcMove
#define RenID		mcRenID
#define Size		mcSize

#define update_ren	mc_update_ren
#define create_ren	mc_create_ren
#define last_stone	mc_last_stone
#define last_ren	mc_last_ren
#define collect_ren	mc_collect_ren
#define remove_ren	mc_remove_ren
#define join_ren	mc_join_ren
#define find_libs_ren	mc_find_libs_ren
#define add_stone	mc_add_stone

#define position	mc_position
#define ren_free	mc_position->ren_free
#if USE_ADJACENT_REN
#define max_ren_free	mc_position->max_ren_free
#endif
#define dirty_ren	mc_dirty_ren
#define dirty_ren0	mc_dirty_ren0
#define captured_stones	mc_captured_stones
#define captured_pos	mc_captured_pos

#define get_color	mc_get_color
#define set_color	mc_set_color
#define clear_color	mc_clear_color
#define is_empty	mc_is_empty
#define is_onboard	mc_is_onboard
#define is_not_ob	mc_is_not_ob
#define is_b_or_w	mc_is_b_or_w
#define is_opponent	mc_is_opponent
#define is_friend	mc_is_friend
#define opposit_color	mc_opposit_color

#define id_ren		mc_id_ren
#define next_ren	mc_next_ren
#define color_ren	mc_color_ren
#define size_ren	mc_size_ren
#define libs_ren	mc_libs_ren
#define member_ren	mc_member_ren
#define liberty_ren	mc_liberty_ren
#define next_member_ren	mc_next_member_ren

#define mark		mc_mark
#define is_marked	mc_is_marked
#define reset_mark	mc_reset_mark
#define mark_ren	mc_mark_ren
#define is_marked_ren	mc_is_marked_ren
#define reset_renmark	mc_reset_renmark
#define empty_list	mc_empty_list
#define remove_from_empty_list	mc_remove_from_empty_list
#if USE_ADJACENT_REN
#define adjacent_ren	mc_adjacent_ren
#endif


#define EMPTY		MC_EMPTY
#define MIN_COLOR	MC_MIN_COLOR
#define NORTH		MC_NORTH
#define EAST		MC_EAST
#define SOUTH		MC_SOUTH
#define WEST		MC_WEST
#define MAX_LIBS_REN	MC_MAX_LIBS_REN
#define FOR_ALL_POINT_IN_REN	MC_FOR_ALL_POINT_IN_REN
#define FOR_4_ADJACENT_POINT	MC_FOR_4_ADJACENT_POINT

#define FATAL		MC_FATAL

#include "update_ren.c"
