/***************************************************************
 * mc.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/************************************************
 *		Monte-Carlo simulator		*
 ************************************************/
#ifndef _MC_H_
#define _MC_H_

#include "config.h"
#include "types.h"
#include "gggo.h"

/*	Constants
 */
#define MC_BOARDSIZE	GG_BOARDSIZE
#define MC_BOARDMIN	GG_BOARDMIN
#define MC_BOARDMAX	GG_BOARDMAX
#define MC_N		GG_N
#define MC_MAX_REN	GG_MAX_REN
#define MC_MAX_LIBS_REN	GG_MAX_LIBS_REN
#define MC_MAX_MOVE	GG_MAX_MOVE
#define MC_MAX_MOVENUM	((MC_MAX_MOVE) * 2)

#define MC_PASS_MOVE	GG_PASS_MOVE
#define MC_NO_MOVE	GG_NO_MOVE
#define MC_RESIGN_MOVE	GG_RESIGN_MOVE

enum {
  MC_N_X = GG_N_X,
  MC_N_Y = GG_N_Y,
};

enum {
  MC_EMPTY = 0,
  MC_WHITE,
  MC_BLACK,
  MC_EXTER,
};

#if USE_SPU
enum { MC_READY = 1, };
enum { MC_NOP = 0, MC_EXEC, MC_EXIT, };
#endif

#define MC_MIN_COLOR		MIN(MC_WHITE, MC_BLACK)
#define MC_MAX_COLOR		MAX(MC_WHITE, MC_BLACK)
#define MC_N_COLOR		(MC_MAX_COLOR - MC_MIN_COLOR + 1)

/*	Types
 */
#define mcBool		ggBool
#define mcColor		ggColor
#define mcRenID		ggRenID
#define mcMove		ggMove
#define mcUint		ggUint
#define mcMark		ggMark
#define mcSize		ggSize
#define mcArc		ggArc

#define mcRen		ggRen
#define mcKoInfo	ggKoInfo
#define mcLink		ggLink

typedef struct {
  uint16_t	n_rave;
  mcMove	move[MC_MAX_MOVENUM];
  mcColor	color[MC_MAX_MOVENUM];
} mcRave		__attribute__((aligned(16)));

typedef struct {
  mcMark	mark_num;
  mcMark	mark_ren_num;
  mcMark	mark[MC_BOARDSIZE];
  uint16_t	bw_captured[MC_N_COLOR];
  uint16_t	movenum;
  mcMove	previous_move;
  mcRenID	ren_free;
  mcRenID	max_ren_free;
  mcKoInfo	ko_info;
  uint8_t	n_pass;
  mcBool	gameend;
  mcColor	color;
#if USE_LADDER_MODE
  mcBool	use_ladder_mode;
#endif
  mcColor	board[MC_BOARDSIZE];
  mcLink	ren_board[MC_BOARDSIZE];
#if USE_EMPTY_LIST
  mcMove	empty_list[MC_BOARDSIZE];
#endif
  mcRen		ren[MC_MAX_REN];
#if USE_ADJACENT_REN
#error Not implemented
  mcBool	adjacent_ren[MC_MAX_REN][MC_MAX_REN];
#endif
} mcPosition		__attribute__((aligned(16)));

typedef struct {
  uint16_t	bw_captured[MC_N_COLOR];
  uint16_t	movenum;
  mcMove	previous_move;
  mcRenID	ren_free;
  mcRenID	max_ren_free;
  mcKoInfo	ko_info;
  uint8_t	n_pass;
  mcBool	gameend;
  mcColor	color;
#if USE_LADDER_MODE
  mcBool	use_ladder_mode;
#endif
  mcColor	board[MC_BOARDSIZE];
  mcLink	ren_board[MC_BOARDSIZE];
#if USE_EMPTY_LIST
  mcMove	empty_list[MC_BOARDSIZE];
#endif
  mcRen		ren[MC_MAX_REN];	/* ren[max_ren_free], in fact */
#if USE_ADJACENT_REN
  mcBool	adjacent_ren[MC_MAX_REN][MC_MAX_REN];
			      /* adjacent_ren[max_ren_free][max_ren_free] */
#endif
} csPackedPosition	__attribute__((aligned(16)));

/*	Macros and Methods
 */
#define MC_NORTH(p)		((p) + MC_N_Y)
#define MC_SOUTH(p)		((p) - MC_N_Y)
#define MC_EAST(p)		((p) + MC_N_X)
#define MC_WEST(p)		((p) - MC_N_X)
#define MC_NE(p)		((p) + MC_N_Y + MC_N_X)
#define MC_NW(p)		((p) + MC_N_Y - MC_N_X)
#define MC_SE(p)		((p) - MC_N_Y + MC_N_X)
#define MC_SW(p)		((p) - MC_N_Y - MC_N_X)
#define MC_NN(p)		((p) + MC_N_Y + MC_N_Y)
#define MC_EE(p)		((p) + MC_N_X + MC_N_X)
#define MC_SS(p)		((p) - MC_N_Y - MC_N_Y)
#define MC_WW(p)		((p) - MC_N_X - MC_N_X)

#define MC_C2X(x)		((x) - MC_MIN_COLOR)

#define MC_FATAL(...) do {		\
  fprintf(stderr, "Abort in %s: ", __func__);	\
  gprintf(__VA_ARGS__);		\
  abort();			\
} while(0)

#define mc_id_ren(p)		mc_position->ren_board[p].id
#define mc_next_member_ren(p)	mc_position->ren_board[p].next
#define mc_color_ren(s)		mc_position->ren[s].color
#define mc_libs_ren(s)		mc_position->ren[s].liberties
#define mc_liberty_ren(s)	mc_position->ren[s].liberty
#define mc_member_ren(s)	mc_position->ren[s].member
#define mc_size_ren(s)		mc_position->ren[s].size
#define mc_next_ren(s)		mc_position->ren[s].next
#if USE_EMPTY_LIST
#define mc_empty_list		mc_position->empty_list
#endif
#if USE_ADJACENT_REN
#define mc_adjacent_ren		mc_position->adjacent_ren
#endif

#ifdef SERVER
extern int		mc_board_size;
#else
#define mc_board_size	gg_board_size
#endif

#define MC_POS(x, y)	((MC_N) * (y) + (x) + (MC_BOARDMIN))
static inline int	MC_X(mcMove p);
static inline int	MC_Y(mcMove p);
static inline int	MC_X(mcMove p) { return (p) % (MC_N) - 1; }
static inline int	MC_Y(mcMove p) { return (p) / (MC_N) - 1; }

extern TLS mcPosition	*mc_position;

extern inline mcColor mc_opposit_color(mcColor c);
extern inline mcColor mc_get_color(unsigned int m);
extern inline mcColor mc_set_color(mcMove m, mcColor c);
extern inline mcColor mc_clear_color(mcMove m);
extern inline mcRenID mc_get_ren_id(mcMove m);
static inline bool mc_is_not_ob(mcMove m);
extern inline bool mc_is_onboard(mcMove m);
static inline bool mc_is_onboard2(mcMove m);
extern inline bool mc_is_empty(mcMove m);
extern inline bool mc_is_exter(mcMove m);
extern inline bool mc_is_b_or_w(mcMove m);
extern inline bool mc_is_friend(mcMove p, mcColor c);
extern inline bool mc_is_opponent(mcMove p, mcColor c);

extern inline mcColor mc_opposit_color(mcColor c)
  { return c ^ (MC_WHITE ^ MC_BLACK); }

extern inline mcColor mc_get_color(unsigned int m)
  { return mc_position->board[m]; }
extern inline mcColor mc_set_color(mcMove m, mcColor c)
  { return mc_position->board[m] = c; }
extern inline mcColor mc_clear_color(mcMove m)
  { return mc_position->board[m] = MC_EMPTY; }
extern inline mcRenID mc_get_ren_id(mcMove m)
  { return mc_position->ren_board[m].id; }

static inline bool mc_is_not_ob(mcMove m)
  { return m < MC_BOARDSIZE; }
extern inline bool mc_is_onboard(mcMove m)
  { return mc_get_color(m) != MC_EXTER; }
static inline bool mc_is_onboard2(mcMove m) {
  return 0 <= MC_X(m) && MC_X(m) < mc_board_size &&
	 0 <= MC_Y(m) && MC_Y(m) < mc_board_size; }
extern inline bool mc_is_empty(mcMove m)
  { return mc_get_color(m) == MC_EMPTY; }
extern inline bool mc_is_exter(mcMove m)
  { return mc_get_color(m) == MC_EXTER; }
extern inline bool mc_is_b_or_w(mcMove m)
  { mcColor c = mc_get_color(m); return c == MC_BLACK || c == MC_WHITE; }
extern inline bool mc_is_friend(mcMove p, mcColor c)
  { return mc_get_color(p) == c; }
extern inline bool mc_is_opponent(mcMove p, mcColor c)
  { return mc_get_color(p) == mc_opposit_color(c); }

static inline char mc_x2c(int i);
static inline char
mc_x2c(int i) { return 'A' + i + ('A' + i >= 'I' ? 'J' - 'I' : 0); }

extern ggScore	mc(mcPosition *p, mcRave *r, mcMove move);

#endif /* _MC_H_ */
