/***************************************************************
 * mc_main.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "mc.h"
#include "queue.h"
#include "cs.h"
#include "pack.h"
#include <string.h>
#ifdef __SPU__
#include <spu_mfcio.h>
#endif

#define gprintf	mc_gprintf
int	mc_gprintf(const char *, ...);

#if PROFILING
#include "timebase.h"
#ifdef __SPU__
static uint32_t profile[GG_N_PROFILE][1];
#define PROFILE_VAR(var)	uint32_t var;
#define PROFILE_CLEAR	\
  for(int i = 0; i < GG_N_PROFILE; i += 1)	\
    for(int j = 0; j < 1; j += 1)	\
      profile[i][j] = 0;
#define PROFILE_START(var)	var = START_CLOCK();
#define PROFILE_RECORD(var, n, id)	profile[n][id] += DELTA_CLOCK(var);
#define PROFILE_RECORD_AND_START(var, n, id)	\
      PROFILE_RECORD(var, n, id)	\
      PROFILE_START(var)
#else
#define PROFILE_VAR(var)
#define PROFILE_CLEAR
#define PROFILE_START(var)
#define PROFILE_RECORD(var, n, id)
#define PROFILE_RECORD_AND_START(var, n, id)
#endif	/* __SPU__ */
#endif /* PROFILING */

extern void	mc_init_gen_rand(uint32_t seed);

static inline ggMessage
mc_pop_msg(csSimulator *p) {
  ggMessage msg;
#ifdef __SPU__
  UNUSED(p);
  msg = spu_read_in_mbox();
#else
  msg = gg_pop(&p->command, NULL);
#endif
  return msg;
}

static void
mc_push_ready_msg(csSimulator *p, ggMessage msg) {
if(0) fprintf(stderr, "--> %s(%p,%#x)\n", __func__, p, msg);
#ifdef __SPU__
  spu_write_out_mbox(msg);	/* spu_writech(SPU_WrOutMbox, msg); */
#else
  gg_push(msg, &p->ready_queue);
#endif
if(0) fprintf(stderr, "<-- %s(%p,%#x)\n", __func__, p, msg);
  return;
}

static void
mc_push_score(csSimulator *p, ggMessage msg) {
if(0) fprintf(stderr, "--> %s(%p,%#x)\n", __func__, p, msg);
#ifdef __SPU__
  spu_write_out_intr_mbox(msg);	/* spu_writech(SPU_WrOutIntrMbox, msg); */
#else
  gg_push(msg, &p->score_queue);
#endif
if(0) fprintf(stderr, "<-- %s(%p,%#x)\n", __func__, p, msg);
  return;
}

static inline void
mc_push_ready(csSimulator *p, int command) {
  mc_push_ready_msg(p, cs_message(0, p->id, command));
  return;
}

static TLS csPacket	cs_packet	__attribute__((aligned(16)));
static TLS csRave	cs_rave		__attribute__((aligned(16)));
static TLS mcPosition	mc_pos		__attribute__((aligned(16)));
static TLS mcRave	mc_rave		__attribute__((aligned(16)));

/************************
 *	MC simulator	*
 ************************/
void	mc_do_simulation(csSimulator *self);
void
mc_do_simulation(csSimulator *self) {
  ggMessage msg;
  ggScore score;
  int command;
  unsigned int size;
  mcMove first_move;
#ifdef __SPU__
  int tag = 0;
#endif

if(0) fprintf(stderr, "--> %s(%p)\n", __func__, self);
  assert(self != NULL);
if(0) fprintf(stderr, "MC[%d] Initialize SFMT seed=%#x\n",
		self->id, self->seed);
  mc_init_gen_rand(self->seed);

  /* Start main loop.
   */
  do {
    mc_push_ready(self, CS_READY);
    msg = mc_pop_msg(self);
if(0) fprintf(stderr, "MC[%d] Read message=%#8x\n", self->id, msg);
    command = cs_command(msg);
    switch(command) {
      default:
	GG_FATAL("Unknown command[%d] %#x\n", self->id, msg);
      /* Neverreached */
      case CS_HALT:
if(2) fprintf(stderr, "MC[%d] Received HALT\n", self->id);
      break;

      case CS_EXEC:
	size = cs_value(msg);
	msg = mc_pop_msg(self);
#ifdef __SPU__
	size = CEIL(size, 16);
	assert(size <= sizeof cs_packet);
if(0) fprintf(stderr, "MC[%d] Copying position dst=%p src=%p size=%#x(%u)\n",
	    self->id, &cs_packet, (void *)(uintptr_t)msg, size, size);
	spu_mfcdma32(&cs_packet, msg, size, tag, MFC_GET_CMD);
	spu_writech(MFC_WrTagMask, 1 << tag);
	(void)spu_mfcstat(MFC_TAG_UPDATE_ALL);
#else
if(0) fprintf(stderr, "MC[%d] Copying position dst=%p src=%p size=%#x(%u)\n",
	    self->id, &cs_packet, (void *)(uintptr_t)msg, size, size);
	memcpy(&cs_packet, (void *)(uintptr_t)msg, size);
#endif
	cs_unpack_position(&mc_pos, &cs_packet.u.position.position);
	first_move = CS_UNPACK(cs_packet.u.position.first_move);
if(0) dump_mc_position(&mc_pos, first_move, stderr);

	score = mc(&mc_pos, &mc_rave, first_move);
if(0) gprintf("MC[%d] Score=%d firstmove=%1m\n", self->id, score, first_move);
	mc_push_score(self, score);

#if USE_RAVE
if(0) dump_mc_rave(&mc_rave, stderr);
	size = cs_pack_rave(&cs_rave, &mc_rave) - (uintptr_t)&cs_rave;
if(0) dump_cs_rave(&cs_rave, stderr);
	mc_push_score(self, size);
	msg = mc_pop_msg(self);
#ifdef __SPU__
	size = CEIL(size, 16);
if(0) fprintf(stderr,
	      "MC[%d] Starting DMA for RAVE src=%p dst=%p size=%#x(%u)\n",
		self->id, &cs_rave, (void *)msg, size, size);
	spu_mfcdma32(&cs_rave, msg, size, tag, MFC_PUT_CMD);
	spu_writech(MFC_WrTagMask, 1 << tag);
	(void)spu_mfcstat(MFC_TAG_UPDATE_ALL);
if(0) fprintf(stderr, "MC[%d] End DMA for RAVE.\n", self->id);
#else
if(0) fprintf(stderr,
	      "MC[%d] Starting DMA for RAVE src=%p dst=%p size=%#x(%u)\n",
		self->id, &cs_rave, (void *)msg, size, size);
	memcpy((void *)msg, &cs_rave, size);
#endif
	mc_push_score(self, CS_DONE);
#endif	/* USE_RAVE */
#if 0
	msg = mc_pop_msg(self);		/* Handshake */
	if(msg != CS_DONE) {
	  MC_FATAL("MC[%d]: Invalid response, msg=%#x\n", self->id, msg);
	}
#endif
      break;
    }
if(0) fprintf(stderr, "MC[%d] Done\n", self->id);
  } while(command != CS_HALT);
if(0) fprintf(stderr, "<-- %s[%d].\n", __func__, self->id);
  return;
}

#ifdef __SPU__
int mc_board_size = GG_MAX_BOARD;	/* TODO Dirty hack */
int
main(uint64_t spe_id, uint64_t id, uint64_t seed) {
  csSimulator cs_simulator;

if(0) fprintf(stderr, "--> SPE_%s(%llu,%#llx)\n", __func__, spe_id, seed);
  cs_simulator.id = (int)id;
  cs_simulator.seed = (uint32_t)seed;
  mc_do_simulation(&cs_simulator);
if(0){fprintf(stderr, "<-- SPE_%s[%llu]\n", __func__, spe_id);fflush(stderr);}
  return 0;
}
#endif
