/***************************************************************
 * queue.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <assert.h>
#include "queue.h"

/****************
 *	Queue	*
 ****************/
int
gg_init_queue(ggQueue *queue) {
  int rc;
if(0) fprintf(stderr, "--> %s(%p)\n", __func__, queue);
  assert(queue != NULL);
  queue->head = queue->tail = queue->len = 0;
  if((rc = pthread_mutex_init(&queue->mutex, NULL)) != 0 ||
     (rc = pthread_cond_init(&queue->cond_not_full, NULL)) != 0 ||
     (rc = pthread_cond_init(&queue->cond_not_empty, NULL)) != 0) {
    perror(__func__);
  }
if(0) fprintf(stderr, "<-- %s rc=%d\n", __func__, rc);
  return rc;
}

int
gg_push(ggMessage msg, ggQueue *queue) {
  int rc;
  bool in_lock;
if(0) fprintf(stderr, "--> %s(%#x,%p)\n", __func__, msg, queue);
  assert(queue != NULL);
  in_lock = false;
  if((rc = pthread_mutex_lock(&queue->mutex)) != 0)	goto error;
  in_lock = true;
  while(queue->len >= GG_MAX_QUEUE_LENGTH) {
    if((rc = pthread_cond_wait(&queue->cond_not_full, &queue->mutex)) != 0)
      goto error;
  }
  queue->data[queue->tail++] = msg;
  if(queue->tail >= GG_MAX_QUEUE_LENGTH)	queue->tail = 0;
  queue->len += 1;
  if((rc = pthread_mutex_unlock(&queue->mutex)) != 0)	goto error;
  in_lock = false;
  if((rc = pthread_cond_signal(&queue->cond_not_empty)) == 0)	goto exit;
error:
  perror(__func__);
  if(in_lock) {
    pthread_mutex_unlock(&queue->mutex);
  }
exit:
if(0) fprintf(stderr, "<-- %s rc=%d\n", __func__, rc);
  return rc;
}

ggMessage
gg_pop(ggQueue *queue, int *rc0) {
  ggMessage msg;
  int rc;
if(0) fprintf(stderr, "--> %s(%p)\n", __func__, queue);
  assert(queue != NULL);
  if((rc = pthread_mutex_lock(&queue->mutex)) != 0)	goto error;
  while(queue->len <= 0) {
    if((rc = pthread_cond_wait(&queue->cond_not_empty, &queue->mutex)) != 0)
      goto error_in_lock;
  }
  msg = queue->data[queue->head++];
  if(queue->head >= GG_MAX_QUEUE_LENGTH)	queue->head = 0;
  queue->len -= 1;
  if((rc = pthread_mutex_unlock(&queue->mutex)) != 0)	goto error;
  if((rc = pthread_cond_signal(&queue->cond_not_full)) == 0) goto exit;
error:
  perror(__func__);
  goto error2;
error_in_lock:
  perror(__func__);
  pthread_mutex_unlock(&queue->mutex);
  goto error2;
error2:
  msg = 0;
  goto exit;
exit:
  if(rc0 != NULL)	*rc0 = rc;
if(0) fprintf(stderr, "<-- %s msg=%#x rc=%d\n", __func__, msg, rc);
  return msg;
}

bool
gg_queue_is_empty(ggQueue *queue, int *rc0) {
  bool empty;
  int rc;
if(0) fprintf(stderr, "--> %s(%p,%p)\n", __func__, queue, rc0);
  assert(queue != NULL);
#if 0
  if((rc = pthread_mutex_lock(&queue->mutex)) == 0) {
    empty = (queue->len == 0);
    if((rc = pthread_mutex_unlock(&queue->mutex)) == 0)	goto exit;
  }
  perror(__func__);
  empty = false;
exit:
#else
  empty = (queue->len == 0);
  rc = 0;
#endif
  if(rc0 != NULL)	*rc0 = rc;
if(0) fprintf(stderr, "<-- %s e=%d rc=%d\n", __func__, empty, rc);
  return empty;
}

bool
gg_queue_is_full(ggQueue *queue, int *rc0) {
  bool full;
  int rc;
  assert(queue != NULL);
if(0) fprintf(stderr, "--> %s(%p,%p)\n", __func__, queue, rc0);
#if 0
  if((rc = pthread_mutex_lock(&queue->mutex)) == 0) {
    full = (queue->len == GG_MAX_QUEUE_LENGTH);
    if((rc = pthread_mutex_unlock(&queue->mutex)) == 0)	goto exit;
  }
  perror(__func__);
  full = false;
exit:
#else
  full = (queue->len == GG_MAX_QUEUE_LENGTH);
  rc = 0;
#endif
  if(rc0 != NULL)	*rc0 = rc;
if(0) fprintf(stderr, "<-- %s e=%d rc=%d\n", __func__, full, rc);
  return full;
}

int
gg_queue_length(ggQueue *queue, int *rc0) {
  int l, rc;
if(0) fprintf(stderr, "--> %s(%p,%p)\n", __func__, queue, rc0);
  assert(queue != NULL);
  if((rc = pthread_mutex_lock(&queue->mutex)) != 0)	goto error;
  l = queue->len;
  if((rc = pthread_mutex_unlock(&queue->mutex)) == 0)	goto exit;
error:
  perror(__func__);
  l = -1;
exit:
  if(rc0 != NULL)	*rc0 = rc;
if(0) fprintf(stderr, "<-- %s l=%d rc=%d\n", __func__, l, rc);
  return l;
}

int
gg_clear_queue(ggQueue *queue) {
  int rc;

if(0) fprintf(stderr, "--> %s(%p)\n", __func__, queue);
  if((rc = pthread_mutex_lock(&queue->mutex)) == 0) {
    queue->head = queue->tail = queue->len = 0;
    if((rc = pthread_cond_init(&queue->cond_not_full, NULL)) != 0) {
      perror(__func__);
    }
    if((rc = pthread_mutex_unlock(&queue->mutex)) == 0)	goto exit;
  }
  perror(__func__);
exit:
if(0) fprintf(stderr, "<-- %s rc=%d\n", __func__, rc);
  return rc;
}
