/***************************************************************
 * server.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/************************************************
 *	Client-Server style implementation	*
 *	     Parallel MCTS Algorithm		*
 *	   over Network using Broadcast		*
 ************************************************/
#include <sys/time.h>		/* gettimeofday() */
#include <time.h>
#include <sys/types.h>		/* getpid() */
#include <unistd.h>
#include <stdlib.h>		/* random() and srandom() */
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include "config.h"
#include "types.h"
#include "mc.h"
#include "queue.h"
#include "uct.h"
#include "cs.h"
#include "pack.h"
#if USE_SPU
#include <libspe2.h>
#endif

#define USE_COPY_ON_SEND	1

/****************
 *	SFMT	*
 ****************/
#define init_gen_rand	gg_init_gen_rand
#define genrand_real2	gg_random
#define gen_rand32	gg_gen_rand32
#define gen_rand64	gg_gen_rand64
// #define get_idstring	gg_get_idstring
#define get_min_array_size32	gg_get_min_array_size32
#define get_min_array_size64	gg_get_min_array_size64
#define fill_array32	gg_fill_array32
#define fill_array64	gg_fill_array64
#define init_by_array	gg_init_by_array
#include "SFMT.c"

extern void	mc_init_gen_rand(uint32_t seed);

#define BLACK	MC_BLACK
#define WHITE	MC_WHITE
#define EMPTY	MC_EMPTY
#define EXTER	MC_EXTER
#define PASS_MOVE	MC_PASS_MOVE
#define RESIGN_MOVE	MC_RESIGN_MOVE
#define POS	MC_POS
#define X	MC_X
#define Y	MC_Y
#define is_not_ob	mc_is_not_ob
#define is_onboard2	mc_is_onboard2
#define x2c	mc_x2c
#include "printutils.c"

#if PROFILING

#include "timebase.h"
#define GG_N_PROFILE	7
static uint32_t profile[GG_N_PROFILE][GG_MAX_THREAD + 1];
#define PROFILE_VAR(var)	uint32_t var;
#define PROFILE_CLEAR	\
  for(int i = 0; i < GG_N_PROFILE; i += 1)	\
    for(int j = 0; j < GG_MAX_THREAD + 1; j += 1)	\
      profile[i][j] = 0;
#define PROFILE_START(var)	var = START_CLOCK();
#define PROFILE_RECORD(var, n, id)	profile[n][id] += DELTA_CLOCK(var);
#define PROFILE_RECORD_AND_START(var, n, id)	\
      PROFILE_RECORD(var, n, id)	\
      PROFILE_START(var)

/* Print out profile
 */
static void
gg_print_profile(uint32_t profile[GG_N_PROFILE][GG_MAX_THREAD + 1]) {

  for(int i = 0; i < GG_MAX_THREAD + 1; i += 1) {
    fprintf(stderr, "Profile[%d]", i);
    for(int j = 0; j < GG_N_PROFILE; j += 1) {
      fprintf(stderr, " %.2fs", SECOND(profile[j][i]));
    }
    gputc('\n');
  }
  fprintf(stderr, "Average:");
  for(int j = 0; j < GG_N_PROFILE; j += 1) {
    float t = 0;
    int percent;
    for(int i = 1; i < GG_MAX_THREAD + 1; i += 1) {
      t += SECOND(profile[j][i]);
    }
    t /= (float)GG_MAX_THREAD;
    percent = roundf(1e2 * t / SECOND(profile[0][0]));
    fprintf(stderr, " %.2fs(%d%%)", t, percent);
  }
  gputc('\n');

#if PRINT_STATISTICS
  for(int i = 0; i < GG_MAX_THREAD + 1; i += 1) {
    fprintf(gg_report, "Profile[%d]", i);
    for(int j = 0; j < GG_N_PROFILE; j += 1) {
      fprintf(gg_report, " %.2fs", SECOND(profile[j][i]));
    }
    fputc('\n', gg_report);
  }
  fprintf(gg_report, "Average:");
  for(int j = 0; j < GG_N_PROFILE; j += 1) {
    float t = 0;
    int percent;
    for(int i = 1; i < GG_MAX_THREAD + 1; i += 1) {
      t += SECOND(profile[j][i]);
    }
    t /= (float)GG_MAX_THREAD;
    percent = roundf(1e2 * t / SECOND(profile[0][0]));
    fprintf(gg_report, " %.2fs(%d%%)", t, percent);
  }
  fputc('\n', gg_report);
  fflush(gg_report);
#endif
  return;
}

#else	/* PROFILING */

#define PROFILE_VAR(var)
#define PROFILE_CLEAR
#define PROFILE_START(var)
#define PROFILE_RECORD(var, n, id)
#define PROFILE_RECORD_AND_START(var, n, id)

#endif	/* PROFILING */

/************************
 *	MC thread 	*
 ************************/
int	mc_board_size = GG_MAX_BOARD;

typedef struct {
  int	recv_fd;
  struct sockaddr_in	recv_sa, from_sa;
} csDatagram;

static int	cs_n_thread;	/* Number of simulation thread */
static const int	yes = 1;

static int
cs_send_packet(csPacket *p, csDatagram *dg, int size) {
  int l;
#if USE_COPY_ON_SEND
  csPacket packet;
#endif

  p->size = CS_PACK((uint16_t)size);
  p->version = CS_PACK(CS_PACKET_VERSION);
  p->type = CS_PACK(CS_PACKET_RESULT);
#if USE_COPY_ON_SEND
  memcpy(&packet, p, size);
  while((l = sendto(dg->recv_fd, &packet, size, MSG_DONTWAIT,
	  (struct sockaddr *)&dg->recv_sa, sizeof dg->recv_sa)) != size)
#else
  while((l = sendto(dg->recv_fd, p, size, 0,
	  (struct sockaddr *)&dg->recv_sa, sizeof dg->recv_sa)) != size)
#endif
  {
    if( ! (l < 0 && errno == EAGAIN)) {
      perror("Server: sendto");
    }
    cs_usleep(5);
  }
if(0) fprintf(stderr, "Server: Sent %3d bytes gen/node/po=%d/%d/%d"
		      " score=%3hd t=%d+%d at %u\n",
	l, CS_UNPACK(p->generation), CS_UNPACK(p->node),
	CS_UNPACK(p->playout), CS_UNPACK(p->u.result.score),
	cs_timesub(CS_UNPACK(p->t_recv), CS_UNPACK(p->t_send)),
	cs_timesub(CS_UNPACK(p->t_done), CS_UNPACK(p->t_recv)),
	CS_UNPACK(p->t_send));
  return l;
}

static int
cs_receive_packet(csPacket *p, csDatagram *dg, uint32_t *time0) {
  int l, size;
  socklen_t sa_size;
  uint32_t time;
  int wrong = 0;

if(0) fprintf(stderr, "--> %s(%p,%p,%p)\n", __func__, p, dg, time0);
  size = 0;
  do {
    sa_size = sizeof dg->recv_sa;
    l = recvfrom(dg->recv_fd, p, sizeof *p, (size == 0) ? 0 : MSG_DONTWAIT,
		    (struct sockaddr *)&dg->recv_sa, &sa_size);
    time = cs_timestamp();
    if(l > 0 && (unsigned)l == CS_UNPACK(p->size) &&
    	CS_UNPACK(p->version) == CS_PACKET_VERSION &&
    	CS_UNPACK(p->type) == CS_PACKET_POSITION) {
      size = l;
      if(time0 != NULL)	*time0 = time;
if(0) fprintf(stderr, "Server: Rcvd %3d bytes gen/node/po=%d/%d/%d @ %u %u"
#if 0
		  " from %s:%u"
#endif
		  "\n",
	  l, CS_UNPACK(p->generation), CS_UNPACK(p->node),
	  CS_UNPACK(p->playout), CS_UNPACK(p->t_send), time
#if 0
	  , inet_ntoa(dg->recv_sa.sin_addr), ntohs(dg->recv_sa.sin_port)
#endif
	  );
    } else if(l < 0) {
      if(errno == EAGAIN)	break;
      perror("Server: recvfrom");
    } else {
      wrong += 1;
if(0) fprintf(stderr, "Server: Rcvd WRONG packet %3d bytes"
		" gen/node/po=%d/%d/%d @ %u %u"
#if 0
		" from %s:%u"
#endif
		"\n",
	l, CS_UNPACK(p->generation), CS_UNPACK(p->node),
	CS_UNPACK(p->playout), CS_UNPACK(p->t_send), time
#if 0
	, inet_ntoa(dg->recv_sa.sin_addr), ntohs(dg->recv_sa.sin_port)
#endif
	);
    }
  } while(true);
if(0) fprintf(stderr, "<-- %s(%p,%p,%p) wrong=%d\n", __func__,
				  p, dg, time0, wrong);
  return wrong;
}

static csDatagram	cs_datagram;
struct {
  csPacket	packet;
  uint8_t	flag;
} cs_packet_pool[GG_MAX_THREAD];
static csSimulator	cs_simulator[GG_MAX_THREAD];

#if USE_SPU

/****************************************
 *	Read from and write to SPU	*
 ****************************************/
static ggMessage
gg_read_ready_from_spe(spe_context_ptr_t spe) {
  int rc;
  uint32_t msg;

if(0) fprintf(stderr, "--> %s(%p)\n", __func__, spe);
  assert(spe != NULL);
  do {
    rc = spe_out_mbox_read(spe, &msg, 1);
    if(rc < 0) {
      if(0) if(errno == EAGAIN)		continue;
      perror(__func__);
      fprintf(stderr, "Error: Reading messages from SPE[%p], rc=%d\n", spe, rc);
      cs_usleep(5);
    }
  } while(rc <= 0);
if(0) fprintf(stderr, "<-- %s %#x rc=%d\n", __func__, msg, rc);
  return (ggMessage)msg;
}

static ggMessage
gg_read_score_from_spe(spe_context_ptr_t spe) {
  int rc;
  uint32_t msg;

if(0) fprintf(stderr, "--> %s(%p)\n", __func__, spe);
  assert(spe != NULL);
  do {
    rc = spe_out_intr_mbox_read(spe, &msg, 1, SPE_MBOX_ALL_BLOCKING);
    if(rc < 0) {
      if(0) if(errno == EAGAIN)		continue;
      perror(__func__);
      fprintf(stderr, "Error: Reading messages from SPE[%p], rc=%d\n", spe, rc);
      cs_usleep(5);
    }
  } while(rc <= 0);
if(0) fprintf(stderr, "<-- %s %#x rc=%d\n", __func__, msg, rc);
  return (ggMessage)msg;
}

static void
gg_write_to_spe(spe_context_ptr_t spe, ggMessage msg) {
  int rc;

if(0) fprintf(stderr, "--> %s(%p,%#x)\n", __func__, spe, msg);
  assert(spe != NULL);
  do {
    rc = spe_in_mbox_write(spe, (uint32_t *)&msg, 1, SPE_MBOX_ALL_BLOCKING);
    if(rc < 0) {
      perror(__func__);
      fprintf(stderr, "Error: Writing message to SPE[%p], rc=%d\n", spe, rc);
//      cs_usleep(5);
    }
  } while(rc <= 0);
if(0) fprintf(stderr, "<-- %s(%p,%#x) rc=%d\n", __func__, spe, msg, rc);
  return;
}

/************************************************
 *	Management thread for SPE		*
 *	Just load and run MC simulator.		*
 ************************************************/
static void
mc_run_spe(csSimulator *self) {
  uint32_t		spe_entry;
  spe_stop_info_t	spe_stop_info;
  int rc;
extern spe_program_handle_t	spe_main_handle;
#define SPE_RUN_FLAGS	0

  rc = spe_program_load(self->spe, &spe_main_handle);
  if(rc != 0) {
    perror(__func__);
    GG_FATAL("Failed to load spe program[%d]\n", self->id);
  }
if(0) fprintf(stderr, "Starting SPE[%d]...\n", self->id);
  spe_entry = SPE_DEFAULT_ENTRY;
  rc = spe_context_run(self->spe, &spe_entry, SPE_RUN_FLAGS,
      (void *)(uint32_t)self->id,
      (void *)self->seed,
      &spe_stop_info);
  if(rc != 0) {
    if(rc < 0) {
      perror(__func__);
      fprintf(stderr, "Failed to run spe program[%d] rc=%d\n", self->id, rc);
    } else {
      fprintf(stderr, "Stopped SPE[%d] rc=%d\n", self->id, rc);
    }
    rc = spe_context_destroy(self->spe);
    if(rc != 0) {
      fprintf(stderr, "context_destroy");
    }
  }
if(0) fprintf(stderr, "Exited SPE[%d] reason=%d result=%#x status=%#x\n",
	  self->id, spe_stop_info.stop_reason,
	  spe_stop_info.result.spe_exit_code, spe_stop_info.spu_status);
  pthread_exit((void *)spe_stop_info.stop_reason);
}

#endif	/* USE_SPU */

static inline void
cs_write_to_mc(csSimulator *p, ggMessage msg) {
#if USE_SPU
  gg_write_to_spe(p->spe, msg);
#else
  gg_push(msg, &p->command);
#endif
  return;
}

static inline void
cs_write_address_to_mc(csSimulator *p, void *q) {
#if USE_SPU
  gg_write_to_spe(p->spe, (uintptr_t)q);
#else
  gg_push((uintptr_t)q, &p->command);
#endif
  return;
}

static inline ggMessage
cs_read_ready(csSimulator *p) {
  assert(p != NULL);
  return
#if USE_SPU
    gg_read_ready_from_spe(p->spe);
#else
    gg_pop(&p->ready_queue, NULL);
#endif
}

static inline ggMessage
cs_read_score(csSimulator *p) {
  assert(p != NULL);
#if USE_SPU
  assert(p->spe != NULL);
  return gg_read_score_from_spe(p->spe);
#else
  return gg_pop(&p->score_queue, NULL);
#endif
}

#if 0
/* Wait completion of transfer. */
static ggMessage
cs_wait_done(csSimulator *p) {
  ggMessage msg = cs_read_ready(p);
  if(cs_command(msg) != CS_DONE) {
    GG_FATAL("Server: invalid response. msg=%#x\n", msg);
  }
  return msg;
}

/* Wait completion of transfer. */
static ggMessage
cs_wait_done_score(csSimulator *p) {
  ggMessage msg = cs_read_score(p);
  if(cs_command(msg) != CS_DONE) {
    GG_FATAL("Server: invalid response. msg=%#x\n", msg);
  }
  return msg;
}
#endif

static inline bool
cs_is_empty_ready(csSimulator *p) {
  return 
#if USE_SPU
    spe_out_mbox_status(p->spe) == 0;
#else
#if 1
    gg_queue_length(&p->ready_queue, NULL) <= 0;
#else
    gg_queue_is_empty(&p->ready_queue, NULL);
#endif
#endif
}

static inline bool
cs_is_empty_score(csSimulator *p) {
  return 
#if USE_SPU
    spe_out_intr_mbox_status(p->spe) == 0;
#else
#if 1
    gg_queue_length(&p->score_queue, NULL) <= 0;
#else
    gg_queue_is_empty(&p->score_queue, NULL);
#endif
#endif
}

/* Poll request from clients.
 */
static int	cs_next_ready, cs_next_score;

static void
cs_init_poll(void) {
  cs_next_ready = 0;
  cs_next_score = 0;
  return;
}

static csSimulator *
cs_poll_ready(void) {
  csSimulator *p;
if(0) fprintf(stderr, "--> %s() next=%d\n", __func__, cs_next_ready);
  for(int i = 0; i < cs_n_thread; i += 1) {
    if(cs_next_ready >= cs_n_thread) cs_next_ready = 0;
    p = &cs_simulator[cs_next_ready++];
    if(!cs_is_empty_ready(p))	goto exit;
  }
  p = NULL;
exit:
if(0) fprintf(stderr, "<-- %s() p=%p next=%d\n", __func__, p, cs_next_ready);
  return p;
}

static csSimulator *
cs_poll_score(void) {
  csSimulator *p;
if(0) fprintf(stderr, "--> %s()\n", __func__);
  for(int i = 0; i < cs_n_thread; i += 1) {
    if(cs_next_score >= cs_n_thread) cs_next_score = 0;
    p = &cs_simulator[cs_next_score++];
    if(!cs_is_empty_score(p))	goto exit;
  }
  p = NULL;
exit:
if(0) fprintf(stderr, "<-- %s() p=%p\n", __func__, p);
  return p;
}

static int	cs_n_packet = -1;
static int
cs_poll_packet(void) {
  int i;
if(0) fprintf(stderr, "--> %s()\n", __func__);
  for(i = 0; i < cs_n_thread; i += 1) {
    if(++cs_n_packet >= cs_n_thread)	cs_n_packet = 0;
    if(cs_packet_pool[cs_n_packet].flag == 0) {
      i = cs_n_packet;
      goto exit;
    }
  }
  i = -1;
exit:
if(0) fprintf(stderr, "<-- %s() i=%d\n", __func__, i);
  return i;
}

/* Main loop.
 * Receive a (position, move), simulate it and send-back the result.
 */
static void
mainloop(void) {
  csSimulator *thread;
  csPacket *p;
  uint32_t msg;
  int size;
  uint32_t t_recv = 0, t_done;
  ggScore score;
  int wrong;
  int packet;

if(0) fprintf(stderr, "--> %s()\n", __func__);
  cs_init_poll();
  for(int i = 0; i < cs_n_thread; i += 1) {
    cs_packet_pool[i].flag = 0;
    cs_simulator[i].packet = 0;
  }
  cs_n_packet = -1;

  do {
    packet = cs_poll_packet();
    if(packet >= 0) {
      p = &cs_packet_pool[packet].packet;
      cs_packet_pool[packet].flag |= CS_IN_USE;
      wrong = cs_receive_packet(p, &cs_datagram, &t_recv);
      mc_board_size = CS_UNPACK(p->u.position.board_size);
if(0) fprintf(stderr,
	  "Server: recv packet gen/node/po=%d/%d/%d arc=%d packet=%d\n",
	  CS_UNPACK(p->generation), CS_UNPACK(p->node),
	  CS_UNPACK(p->playout), CS_UNPACK(p->optarc), packet);
      /* Other fields have no need to unpack because
       * they will be sent back as are for the use by update_tree().
       */
      size = (int)CS_UNPACK(p->size);

      /* Check a request from mc_simulating threads.
       */
      if((thread = cs_poll_ready()) != NULL) {
if(0) fprintf(stderr, "Server: MC[%d] is ready, packet=%d\n",
		  thread->id, packet);
	thread->packet = packet;
	msg = cs_read_ready(thread);
	cs_write_to_mc(thread, cs_message(size, 0, CS_EXEC));
	cs_write_address_to_mc(thread, p);
if(0) fprintf(stderr, "Server: MC[%d](%p) start\n", thread->id, thread);
      }
    }

    if((thread = cs_poll_score()) != NULL) {
      packet = thread->packet;
      p = &cs_packet_pool[packet].packet;
      score = cs_read_score(thread);
if(0) fprintf(stderr, "Server: MC[%d] end, score=%hd, packet=%d\n",
		  thread->id, score, packet);
#if USE_RAVE
      size = cs_read_score(thread);
if(0) fprintf(stderr, "Server: MC[%d] size for rave=%d\n", thread->id, size);
      /* Caution! This works only for 32 bit applications. */
      cs_write_to_mc(thread, (uintptr_t)&p->u.result.rave);
      msg = cs_read_score(thread);
      if(cs_command(msg) != CS_DONE) {
	GG_FATAL("Server: invalid response. thread=%p msg=%#x\n", thread, msg);
      }
if(0) dump_cs_rave(&p->u.result.rave, stderr);
#endif
#if 0
      cs_write_to_mc(thread, cs_message(0, 0, CS_DONE));
#endif

      /* Sendback the result.
       */
if(0) fprintf(stderr, "Server: MC[%d] start sending\n", thread->id);
      size = (uintptr_t)&p->u.result.rave + size - (uintptr_t)p;
      t_done = cs_timestamp();
      p->u.result.score = CS_PACK(score);
      p->t_done = CS_PACK(t_done);
      cs_send_packet(p, &cs_datagram, size);
      cs_packet_pool[packet].flag &= ~CS_IN_USE;
if(0) fprintf(stderr, "Server: MC[%d] sent, size=%d\n", thread->id, size);
    }
  } while(true);
if(0) fprintf(stderr, "<-- %s\n", __func__);
  return;
}

int
main(int argc, char *argv[]) {
  unsigned int seed;
  uint16_t port;
  int rc;
  int i;

  port = CS_PORT;
{
  struct timeval tv;
  (void)gettimeofday(&tv, NULL);
  seed = (ggRandomSeed)(tv.tv_sec * 1000000 + tv.tv_usec + getpid());
}
  cs_n_thread = GG_MAX_THREAD;
  switch(argc) {
    default:
      fprintf(stderr, "Usage: %s [n_thread=%d [port=%d seed=%#x]]\n",
      	  argv[0], cs_n_thread, port, seed);
      exit(EXIT_FAILURE);
    case 4:
      seed = atol(argv[3]);
    case 3:
      port = atoi(argv[2]);
    case 2:
      cs_n_thread = atoi(argv[1]);
    case 1:
    break;
  }
  if(cs_n_thread > GG_MAX_THREAD) {
    fprintf(stderr, "Too many threads. Upto %d\n", GG_MAX_THREAD);
    exit(EXIT_FAILURE);
  }

  srandom(seed);

  /* Setup sockets.
   */
  if((cs_datagram.recv_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    cs_exit_perror("Server: socket");
  }
  if(setsockopt(cs_datagram.recv_fd, SOL_SOCKET,
		    SO_REUSEADDR, &yes, sizeof yes) < 0) {
    cs_exit_perror("Server: setsockopt");
  }
  memset(&cs_datagram.recv_sa, 0, sizeof cs_datagram.recv_sa);
  cs_datagram.recv_sa.sin_family = AF_INET;
  cs_datagram.recv_sa.sin_addr.s_addr = htonl(INADDR_ANY);
  cs_datagram.recv_sa.sin_port = htons(port);
  if(bind(cs_datagram.recv_fd, (struct sockaddr *)&cs_datagram.recv_sa,
					sizeof cs_datagram.recv_sa) < 0) {
    cs_exit_perror("Server: bind");
  }
  memset(&cs_datagram.from_sa, 0, sizeof cs_datagram.from_sa);

  /* Initialize server thread control block.
   */
  for(i = 0; i < cs_n_thread; i += 1) {
    cs_simulator[i].id = i;
    cs_simulator[i].seed = random();
#if ! USE_SPU
    gg_init_queue(&cs_simulator[i].command);
    gg_init_queue(&cs_simulator[i].ready_queue);
    gg_init_queue(&cs_simulator[i].score_queue);
#endif
    /* Create server-side simulation thread. */
#if USE_SPU
    cs_simulator[i].spe = spe_context_create(0, NULL);
    if(cs_simulator[i].spe == NULL) {
      perror("context_create");
      GG_FATAL("Failed to create spe context[%d].\n", i);
    }
    rc = pthread_create(&cs_simulator[i].thread, NULL,
	    (void *(*)(void *))mc_run_spe, (void *)&cs_simulator[i]);
#else
    rc = pthread_create(&cs_simulator[i].thread, NULL,
	    (void *(*)(void *))mc_do_simulation, (void *)&cs_simulator[i]);
#endif
    if(rc != 0) {
      perror(__func__);
      GG_FATAL("Can't create thread[%d] rc=%d -- %s\n", i, rc, __func__);
    }
  }
  cs_usleep(10);
  mainloop();
  close(cs_datagram.recv_fd);
  return 0;
}
