/***************************************************************
 * simple_server.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/************************************************
 *	Client-Server style implementation	*
 *	  Parallel MCTS Algorithm		*
 *	  over Network using UDP/IP		*
 ************************************************/
#define USE_SIGIO	1

#include <sys/time.h>	/* For gettimeofday() */
#include <time.h>
#include <sys/types.h>	/* For getpid() */
#include <unistd.h>
#include <stdlib.h>	/* For random() and srandom() */
#if USE_SIGIO
#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#endif
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include "config.h"
#include "types.h"
#include "mc.h"
#include "queue.h"
#include "uct.h"
#include "cs.h"
#include "pack.h"

/****************
 *	SFMT	*
 ****************/
#define init_gen_rand	gg_init_gen_rand
#define genrand_real2	gg_random
#define gen_rand32	gg_gen_rand32
#define gen_rand64	gg_gen_rand64
// #define get_idstring	gg_get_idstring
#define get_min_array_size32	gg_get_min_array_size32
#define get_min_array_size64	gg_get_min_array_size64
#define fill_array32	gg_fill_array32
#define fill_array64	gg_fill_array64
#define init_by_array	gg_init_by_array
#include "SFMT.c"

extern void	mc_init_gen_rand(uint32_t seed);

#define BLACK	MC_BLACK
#define WHITE	MC_WHITE
#define EMPTY	MC_EMPTY
#define EXTER	MC_EXTER
#define PASS_MOVE	MC_PASS_MOVE
#define RESIGN_MOVE	MC_RESIGN_MOVE
#define POS	MC_POS
#define X	MC_X
#define Y	MC_Y
#define is_not_ob	mc_is_not_ob
#define is_onboard2	mc_is_onboard2
#define x2c	mc_x2c
#include "printutils.c"

#if PROFILING

#include "timebase.h"
#define GG_N_PROFILE	7
static uint32_t profile[GG_N_PROFILE][GG_MAX_THREAD + 1];
#define PROFILE_VAR(var)	uint32_t var;
#define PROFILE_CLEAR	\
  for(int i = 0; i < GG_N_PROFILE; i += 1)	\
    for(int j = 0; j < GG_MAX_THREAD + 1; j += 1)	\
      profile[i][j] = 0;
#define PROFILE_START(var)	var = START_CLOCK();
#define PROFILE_RECORD(var, n, id)	profile[n][id] += DELTA_CLOCK(var);
#define PROFILE_RECORD_AND_START(var, n, id)	\
      PROFILE_RECORD(var, n, id)	\
      PROFILE_START(var)

/* Print out profile
 */
static void
gg_print_profile(uint32_t profile[GG_N_PROFILE][GG_MAX_THREAD + 1]) {

  for(int i = 0; i < GG_MAX_THREAD + 1; i += 1) {
    gprintf("Profile[%d]", i);
    for(int j = 0; j < GG_N_PROFILE; j += 1) {
      gprintf(" %.2fs", SECOND(profile[j][i]));
    }
    gputc('\n');
  }
  gprintf("Average:");
  for(int j = 0; j < GG_N_PROFILE; j += 1) {
    float t = 0;
    int percent;
    for(int i = 1; i < GG_MAX_THREAD + 1; i += 1) {
      t += SECOND(profile[j][i]);
    }
    t /= (float)GG_MAX_THREAD;
    percent = roundf(1e2 * t / SECOND(profile[0][0]));
    gprintf(" %.2fs(%d%%)", t, percent);
  }
  gputc('\n');

#if PRINT_STATISTICS
  for(int i = 0; i < GG_MAX_THREAD + 1; i += 1) {
    fprintf(gg_report, "Profile[%d]", i);
    for(int j = 0; j < GG_N_PROFILE; j += 1) {
      fprintf(gg_report, " %.2fs", SECOND(profile[j][i]));
    }
    fputc('\n', gg_report);
  }
  fprintf(gg_report, "Average:");
  for(int j = 0; j < GG_N_PROFILE; j += 1) {
    float t = 0;
    int percent;
    for(int i = 1; i < GG_MAX_THREAD + 1; i += 1) {
      t += SECOND(profile[j][i]);
    }
    t /= (float)GG_MAX_THREAD;
    percent = roundf(1e2 * t / SECOND(profile[0][0]));
    fprintf(gg_report, " %.2fs(%d%%)", t, percent);
  }
  fputc('\n', gg_report);
  fflush(gg_report);
#endif
  return;
}

#else	/* PROFILING */

#define PROFILE_VAR(var)
#define PROFILE_CLEAR
#define PROFILE_START(var)
#define PROFILE_RECORD(var, n, id)
#define PROFILE_RECORD_AND_START(var, n, id)

#endif	/* PROFILING */

extern void	mc_init_gen_rand(uint32_t seed);

int		mc_board_size = GG_MAX_BOARD;

static csPacket		cs_packet;
static mcPosition	cs_position;
#if USE_RAVE
static mcRave		cs_rave;
#endif

static int		cs_recv_fd;
static struct sockaddr_in	cs_recv_sa;
static const int	cs_yes = 1;

#if USE_SIGIO

static volatile int	io_nqueue;
static uint32_t		io_t_recv;
static csPacket		io_packet;

static void
cs_sig_io(int signo) {
  ssize_t len;
  socklen_t	salen;
  struct timeval tv;
  UNUSED(signo);

  do {
    salen = sizeof cs_recv_sa;
    len = recvfrom(cs_recv_fd, (char *)&io_packet, sizeof(csPacket), 0,
		      &cs_recv_sa, &salen);
    if(len < 0) {
      if(errno == EWOULDBLOCK)	break;
      perror("Server: recvfrom");
    }
    if(ioctl(cs_recv_fd, SIOCGSTAMP, &tv) == 0) {
      io_t_recv = tv.tv_usec;
    } else {
      perror("ioctl: SIOCGSTAMP");
      io_t_recv = cs_timestamp();
    }
  } while(true);
  io_nqueue = 1;
  return;
}

static int
cs_ioctl3(int d, int req, const int *val) {
  int rc = ioctl(d, req, val);
  if(rc < 0)	perror(__func__);
  return rc;
}

static sighandler_t
cs_signal(int signum, sighandler_t handler) {
  sighandler_t oldhadler = signal(signum, handler);
  if(oldhadler == SIG_ERR)	perror(__func__);
  return oldhadler;
}

static int
cs_sigemptyset(sigset_t *set) {
  int rc = sigemptyset(set);
  if(rc < 0)	perror(__func__);
  return rc;
}

static int
cs_sigaddset(sigset_t *set, int signum) {
  int rc =sigaddset(set, signum);
  return rc;
}

static int
cs_sigprocmask(int how, const sigset_t *set, sigset_t *oldset) {
  int rc = sigprocmask(how, set, oldset);
  if(rc < 0)	perror(__func__);
  return rc;
}

#else	/* USE_SIGIO */

static int
cs_receive_packet(uint32_t *time) {
  int l, size;
  socklen_t sa_size;
  int wrong = 0;

if(0) fprintf(stderr, "--> %s(%p)\n", __func__, time);
  assert(time != NULL);
  size = 0;
  do {
    sa_size = sizeof cs_recv_sa;
    l = recvfrom(cs_recv_fd, &cs_packet, sizeof cs_packet,
    	(size == 0) ? 0 : MSG_DONTWAIT, (struct sockaddr *)&cs_recv_sa,
    	&sa_size);
    if(l > 0 && (unsigned)l == CS_UNPACK(cs_packet.size) &&
    	CS_UNPACK(cs_packet.version) == 1 && CS_UNPACK(cs_packet.type) == 1) {
      size = l;
      *time = cs_timestamp();
if(0) fprintf(stderr, "Server: Rcvd %3d bytes gen/node/po=%d/%d/%d @ %u %u"
#if 0
		  " from %s:%u"
#endif
		  "\n",
	  l, CS_UNPACK(cs_packet.generation), CS_UNPACK(cs_packet.node),
	  CS_UNPACK(cs_packet.playout), CS_UNPACK(cs_packet.t_send), *time
#if 0
	  , inet_ntoa(cs_recv_sa.sin_addr), ntohs(cs_recv_sa.sin_port)
#endif
	  );
    } else if(l < 0) {
      if(errno == EAGAIN)	break;
      perror("Server: recvfrom");
    } else {
      wrong += 1;
if(0) fprintf(stderr, "Server: Rcvd WRONG packet %3d bytes"
		" gen/node/po=%d/%d/%d @ %u %u from %s:%u\n",
	l, CS_UNPACK(cs_packet.generation), CS_UNPACK(cs_packet.node),
	CS_UNPACK(cs_packet.playout), CS_UNPACK(cs_packet.t_send), *time,
	inet_ntoa(cs_recv_sa.sin_addr), ntohs(cs_recv_sa.sin_port));
    }
  } while(true);
if(0) fprintf(stderr, "<-- %s(%p) l=%d\n", __func__, time, l);
  return wrong;
}

#endif	/* USE_SIGIO */

int
main(int argc, char *argv[]) {
  unsigned int seed;
  uint16_t port;
  int l, size;
  uint32_t t_recv = 0, t_done;
  mcMove first_move;
  ggScore score;
  struct timeval tv;
#if USE_SIGIO
  sigset_t zeromask, newmask, oldmask;
#else
  int wrong;
#endif

  port = CS_PORT;
  (void)gettimeofday(&tv, NULL);
  seed = (ggRandomSeed)(tv.tv_sec * 1000000 + tv.tv_usec + getpid());

  switch(argc) {
    default:
      fprintf(stderr, "Usage: %s [port=%d [seed=%#x]]\n", argv[0], port, seed);
      exit(EXIT_FAILURE);
    case 3:
      seed = atol(argv[3]);
    case 2:
      port = atoi(argv[2]);
    case 1:
    break;
  }

  /* Setup sockets. */
  if((cs_recv_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    cs_exit_perror("Server: socket");
  }
  if(setsockopt(cs_recv_fd, SOL_SOCKET, SO_REUSEADDR, &cs_yes, sizeof cs_yes) < 0) {
    cs_exit_perror("Server: setsockopt");
  }
  memset(&cs_recv_sa, 0, sizeof cs_recv_sa);
  cs_recv_sa.sin_family = AF_INET;
  cs_recv_sa.sin_addr.s_addr = htonl(INADDR_ANY);
  cs_recv_sa.sin_port = htons(port);
  if(bind(cs_recv_fd, (struct sockaddr *)&cs_recv_sa, sizeof cs_recv_sa) < 0) {
    cs_exit_perror("Server: bind");
  }

#if USE_SIGIO
  io_nqueue = 0;
  cs_signal(SIGIO, cs_sig_io);
  if(fcntl(cs_recv_fd, F_SETOWN, getpid()) < 0) {
    cs_exit_perror("Server: fcntl");
  }
  cs_ioctl3(cs_recv_fd, FIOASYNC, &cs_yes);
  cs_ioctl3(cs_recv_fd, FIONBIO, &cs_yes);
  cs_sigemptyset(&zeromask);
  cs_sigemptyset(&oldmask);
  cs_sigemptyset(&newmask);
  cs_sigaddset(&newmask, SIGIO);
#endif

  mc_init_gen_rand(seed);

  /* Main loop.
   * Receive a {position, move}, simulate it until game end,
   * and return the score and optional {color, move}'s for RAVE.
   */

  do {

#if USE_SIGIO
    while(io_nqueue == 0) {
      sigsuspend(&zeromask);
    }
    t_recv = io_t_recv;
    cs_packet = io_packet;
    cs_sigprocmask(SIG_SETMASK, &oldmask, NULL);
    cs_unpack_position(&cs_position, &cs_packet.u.position.position);
#else
    wrong = cs_receive_packet(&t_recv);
    cs_unpack_position(&cs_position, &cs_packet.u.position.position);
#endif
if(0) fprintf(stderr, "Server: Start simulation, gen/node/po=%d/%d/%d arc=%d\n",
	CS_UNPACK(cs_packet.generation), CS_UNPACK(cs_packet.node),
	CS_UNPACK(cs_packet.playout), CS_UNPACK(cs_packet.optarc));
    mc_board_size = CS_UNPACK(cs_packet.u.position.board_size);
    first_move = CS_UNPACK(cs_packet.u.position.first_move);
if(0) dump_mc_position(&cs_position, first_move, stderr);

#if USE_RAVE
    score = mc(&cs_position, &cs_rave, first_move);
if(0) fprintf(stderr, "Server: Score=%d\n", score);
if(0) dump_mc_rave(&cs_rave, stderr);
    size = cs_pack_rave(&cs_packet.u.result.rave, &cs_rave)
		- (uintptr_t)&cs_packet;
if(0) fprintf(stderr, "Server: Rave size=%d\n", size);
if(0) dump_cs_rave(&cs_packet.u.result.rave, stderr);
#else
    score = mc(&cs_position, first_move);
if(0) fprintf(stderr, "Server: Score=%d\n", score);
    size = (uintptr_t)&cs_packet.u.result + sizeof(ggScore)
		- (uintptr_t)&cs_packet;
#endif
if(0) fprintf(stderr, "Server: total size=%d\n", size);

    /* Sendback the result
     */
    t_done = cs_timestamp();
    cs_packet.u.result.score = CS_PACK(score);
    cs_packet.size = CS_PACK((uint16_t)size);
    cs_packet.t_recv = CS_PACK(t_recv);
    cs_packet.t_done = CS_PACK(t_done);
    cs_packet.version = CS_PACK((uint8_t)1);
    cs_packet.type = CS_PACK((uint8_t)2);
    while((l = sendto(cs_recv_fd, &cs_packet, size, MSG_DONTWAIT,
	      (struct sockaddr *)&cs_recv_sa, sizeof cs_recv_sa)) != size) {
      if( ! (l < 0 && errno == EAGAIN)) {
	perror("Server: sendto");
      }
      cs_usleep(1);
    }
if(0) fprintf(stderr, "Server: Sent %3d bytes gen/node/po=%d/%d/%d"
			" score=%3d t=%d+%d at %u\n",
	l, CS_UNPACK(cs_packet.generation), CS_UNPACK(cs_packet.node),
	CS_UNPACK(cs_packet.playout),
	(CS_UNPACK(cs_packet.u.result.score) << 16) >> 16,
	cs_timesub(CS_UNPACK(cs_packet.t_recv), CS_UNPACK(cs_packet.t_send)),
	cs_timesub(CS_UNPACK(cs_packet.t_done), CS_UNPACK(cs_packet.t_recv)),
	CS_UNPACK(cs_packet.t_send));
#if USE_SIGIO
    cs_sigprocmask(SIG_BLOCK, &newmask, &oldmask);
    io_nqueue = 0;
#endif
  } while(true);

if(0) fprintf(stderr, "<-- %s\n", __func__);
  close(cs_recv_fd);
  return 0;
}
